#!/usr/bin/env python3
#  Copyright (c) 2019 Toyota Research Institute

"""
Helper functions for generating features in beep.featurize module
All methods are currently lumped into this script.
"""
import os
import calendar

import pandas as pd
import numpy as np
from scipy.interpolate import interp1d
from scipy.optimize import curve_fit
from scipy.stats import skew, kurtosis


from beep import PROTOCOL_PARAMETERS_DIR
from beep.utils import parameters_lookup


def list_minus(list1, list2):
    """
    this function takes in two lists and will return a list containing
    the values of list1 minus list2
    """
    result = []
    zip_object = zip(list1, list2)
    for list1_i, list2_i in zip_object:
        result.append(list1_i - list2_i)
    return result


def get_hppc_ocv_helper(cycle_hppc_0, step_num):
    """
    this helper function takes in a cycle and a step number
    and returns a list that stores the mean of the last five points of voltage in different
    step counter indexes (which is basically the soc window)
    """
    chosen1 = cycle_hppc_0[cycle_hppc_0.step_index == step_num]
    voltage1 = []
    step_index_counters = chosen1.step_index_counter.unique()[0:9]
    for i in range(len(step_index_counters)):
        df_i = chosen1.loc[chosen1.step_index_counter == step_index_counters[i]]
        voltage1.append(
            df_i["voltage"].iloc[-10].mean()
        )  # take the mean of the last 10 points of the voltage value
    return voltage1


def get_hppc_ocv(processed_cycler_run, diag_pos, parameters_path=PROTOCOL_PARAMETERS_DIR):
    """
    This function calculates the variance, min, mean, skew, kurtosis, sum and sum of squares 
    of ocv changes between hppc cycle specified by and the first one.

    Argument:
            processed_cycler_run (beep.structure.ProcessedCyclerRun)
            diag_pos (int): diagnostic cycle occurence for a specific <diagnostic_cycle_type>. e.g.
            if rpt_0.2C, occurs at cycle_index = [2, 37, 142, 244 ...], <diag_pos>=0 would correspond to cycle_index 2.
    Returns:
            a dataframe with seven entries 
            ('var_ocv, min_ocv, mean_ocv, skew_ocv, kurtosis_ocv, sum_ocv, sum_square_ocv'):
    """

    hppc_ocv_features = pd.DataFrame()

    data = processed_cycler_run.diagnostic_data
    cycle_hppc = data.loc[data.cycle_type == "hppc"]
    cycle_hppc = cycle_hppc.loc[cycle_hppc.current.notna()]
    cycles = cycle_hppc.cycle_index.unique()

    cycle_hppc_0 = cycle_hppc.loc[cycle_hppc.cycle_index == cycles[0]]

    first_diagnostic_steps = get_step_index(processed_cycler_run,
                                            cycle_type="hppc",
                                            diag_pos=0,
                                            parameters_path=parameters_path
                                            )
    later_diagnostic_steps = get_step_index(processed_cycler_run,
                                            cycle_type="hppc",
                                            diag_pos=diag_pos,
                                            parameters_path=parameters_path
                                            )
    step_first = first_diagnostic_steps['hppc_long_rest']
    step_later = later_diagnostic_steps['hppc_long_rest']

    voltage_1 = get_hppc_ocv_helper(cycle_hppc_0, step_first)
    selected_diag_df = cycle_hppc.loc[cycle_hppc.cycle_index == cycles[diag_pos]]
    voltage_2 = get_hppc_ocv_helper(selected_diag_df, step_later)

    ocv = list_minus(voltage_1, voltage_2)

    hppc_ocv_features["var_ocv"] = [np.var(ocv)]
    hppc_ocv_features["min_ocv"] = [min(ocv)]
    hppc_ocv_features["mean_ocv"] = [np.mean(ocv)]
    hppc_ocv_features["skew_ocv"] = [skew(ocv)]
    hppc_ocv_features["kurtosis_ocv"] = [kurtosis(ocv, fisher=False, bias=False)]
    hppc_ocv_features["sum_ocv"] = [np.sum(np.absolute(ocv))]
    hppc_ocv_features["sum_square_ocv"] = [np.sum(np.square(ocv))]

    return hppc_ocv_features


def res_calc(chosen, soc, r_type):

    """
    This function calculates resistance based on different socs and differnet time scales in hppc cycles.
    Args:
        chosen(pd.DataFrame): a dataframe for a specific diagnostic cycle you are interested in.
        soc (int): step index counter corresponding to the soc window of interest - 0, 1, 2, 3, 4 ... 
        r_type (str): a string that indicates the time scale of the resistance you are calculating, e.g. 
        'r_c_0s', 'r_c_3s', 'r_c_end', 'r_d_0s', 'r_d_3s', 'r_d_end'
    Returns:
        charge/discharge resistance value (float) at this specific soc and time scale in hppc cycles
    """
    steps = chosen.step_index.unique()[1:6]
    counters = []
    for step in steps:
        counters.append(chosen[chosen.step_index == step].step_index_counter.unique().tolist())
    # for charge 
    if r_type[2] == 'c':
        # 40 s short rest 
        step_ocv = 2
        step_cur = 3
    # for discharge 
    if r_type[2] == 'd':
        # one hour long rest 
        step_ocv = 0
        step_cur = 1
    # if there is no dataframe for ocv or step cur, it means this step is skipped, so return None directly 
    try:
        chosen_step_ocv = chosen[(chosen.step_index_counter == counters[step_ocv][soc])]
        chosen_step_cur = chosen[chosen.step_index_counter == counters[step_cur][soc]]
    except IndexError:
        return None
    # since the data is voltage interpolated, so we want to sort the data based on time 
    chosen_step_ocv = chosen_step_ocv.sort_values(by='test_time')
    chosen_step_cur = chosen_step_cur.sort_values(by='test_time')
    # last data point of the rest is the ocv value 
    v_ocv = chosen_step_ocv.voltage.iloc[-1]
    # taking the average of the last 5 data points of the current 
    i_ocv = chosen_step_ocv.current.tail(5).mean()
    # now we look at the time scales   
    if r_type[4] == 'e':
        v_dis = chosen_step_cur.voltage.iloc[-1]
        i_dis = chosen_step_cur.current.iloc[-1]
        res = (v_dis - v_ocv)/(i_dis-i_ocv)
        return res
    else:
        if r_type[4] == '0':
            index = 0.001
        elif r_type[4] == '3':
            index = 3
        # test time is in the units of s 
        chosen_step_cur_index = chosen_step_cur[(chosen_step_cur.test_time - chosen_step_cur.test_time.min()) <= index]
        v_dis = chosen_step_cur_index.voltage.iloc[-1]
        i_dis = chosen_step_cur_index.current.iloc[-1]
        res = (v_dis - v_ocv)/(i_dis-i_ocv)
        return res


def get_resistance_soc_duration_hppc(processed_cycler_run, diag_pos):
    """
    This function calculates resistances based on different socs and differnet time scales for a targeted hppc cycle.
    Args:
        processed_cycler_run (beep.structure.ProcessedCyclerRun)
        diag_pos (int): diagnostic cycle occurence for a specific <diagnostic_cycle_type>. e.g.
        if rpt_0.2C, occurs at cycle_index = [2, 37, 142, 244 ...], <diag_pos>=0 would correspond to cycle_index 2
    Returns:
        a dataframe (single row)
        - and its 54 columns list all the possible resistance names 'r_c_0s_0', 'r_c_3s_0'...
            - r: resistance
            - c/d: state (charge or discharge)
            - timescale: 0s, 3s, or end of the cycle (resistance)
            - soc_index: an int indicating which soc window in HPPC e.g. 0, 1, 2,...
    """
    data = processed_cycler_run.diagnostic_data
    hppc_cycle = data.loc[data.cycle_type == 'hppc']
    hppc_cycle = hppc_cycle.loc[hppc_cycle.current.notna()]
    cycles = hppc_cycle.cycle_index.unique()
    # a list of strings to get charge/discharge resistances at different time scales
    names = ['r_c_0s', 'r_c_3s', 'r_c_end', 'r_d_0s', 'r_d_3s', 'r_d_end']
    output = pd.DataFrame()
    chosen = hppc_cycle[hppc_cycle.cycle_index == cycles[diag_pos]]
    # for each diagnostic cycle, we have a row conatins all the resistances 
    df_row = pd.DataFrame()
    for name in names:
        for j in range(9):
            # full name 
            f_name = name + '_' + str(j)
            df_row[f_name] = [res_calc(chosen, j, name)]
    output = output.append(df_row, ignore_index=True)
    return output


def get_dr_df(processed_cycler_run, diag_pos):
    """
    This function calculates resistance changes between a hppc cycle specified by and the first one under different
    pulse durations (1ms for ohmic resistance, 2s for charge transfer and the end of pulse for polarization resistance)
    and different state of charge.
    Args:
        processed_cycler_run (beep.structure.ProcessedCyclerRun)
        diag_pos (int): diagnostic cycle occurence for a specific <diagnostic_cycle_type>. e.g.
        if rpt_0.2C, occurs at cycle_index = [2, 37, 142, 244 ...], <diag_pos>=0 would correspond to cycle_index 2.
    Returns:
        a dataframe contains resistances changes normalized by the first diagnostic cycle value.
    """
    r_df_0 = get_resistance_soc_duration_hppc(processed_cycler_run, 0)
    r_df_i = get_resistance_soc_duration_hppc(processed_cycler_run, diag_pos)
    dr_df = (r_df_i - r_df_0) / r_df_0
    return dr_df


def get_V_I(df):
    """
    this helper functiion takes in a specific step in the first hppc cycle and gives you the voltage values as
    well as the current values after each step in the first cycle.
    """
    result = {}
    voltage = []
    current = []
    step_index_counters = df.step_index_counter.unique()[0:9]
    for i in range(len(step_index_counters)):
        df_i = df.loc[df.step_index_counter == step_index_counters[i]]
        voltage.append(df_i["voltage"].iloc[-1])  # the last point of the voltage value
        current.append(df_i["current"].mean())
    result["voltage"] = voltage
    result["current"] = current
    return result


def get_v_diff(processed_cycler_run, diag_pos, soc_window, parameters_path=PROTOCOL_PARAMETERS_DIR):
    """
    This method calculates the variance of voltage difference between a specified hppc cycle and the first
    one during a specified state of charge window.
    Args:
        processed_cycler_run (beep.structure.ProcessedCyclerRun)
        diag_pos (int): diagnostic cycle occurence for a specific <diagnostic_cycle_type>. e.g.
        if rpt_0.2C, occurs at cycle_index = [2, 37, 142, 244 ...], <diag_pos>=0 would correspond to cycle_index 2
        soc_window (int): step index counter corresponding to the soc window of interest.
    Returns:
        a dataframe that contains the variance of the voltage differences
    """

    result = pd.DataFrame()

    data = processed_cycler_run.diagnostic_data
    hppc_data = data.loc[data.cycle_type == "hppc"]
    cycles = hppc_data.cycle_index.unique()

    hppc_data_2 = hppc_data.loc[hppc_data.cycle_index == cycles[diag_pos]]
    hppc_data_1 = hppc_data.loc[hppc_data.cycle_index == cycles[0]]
    #     in case a final HPPC is appended in the end also with cycle number 2
    hppc_data_1 = hppc_data_1.loc[hppc_data_1.discharge_capacity < 8]

    step_ind_1 = get_step_index(processed_cycler_run, cycle_type="hppc", diag_pos=0, parameters_path=parameters_path)
    step_ind_2 = get_step_index(processed_cycler_run, cycle_type="hppc", diag_pos=1, parameters_path=parameters_path)

    hppc_data_2_d = hppc_data_2.loc[hppc_data_2.step_index == step_ind_2["hppc_discharge_to_next_soc"]]
    hppc_data_1_d = hppc_data_1.loc[hppc_data_1.step_index == step_ind_1["hppc_discharge_to_next_soc"]]
    step_counters_1 = hppc_data_1_d.step_index_counter.unique()
    step_counters_2 = hppc_data_2_d.step_index_counter.unique()

    if min(len(step_counters_1) - 1, len(step_counters_2) - 1) < soc_window:
        return None

    chosen_1 = hppc_data_1_d.loc[hppc_data_1_d.step_index_counter == step_counters_1[soc_window]]
    chosen_2 = hppc_data_2_d.loc[hppc_data_2_d.step_index_counter == step_counters_2[soc_window]]
    chosen_1 = chosen_1.loc[chosen_1.discharge_capacity.notna()]
    chosen_2 = chosen_2.loc[chosen_2.discharge_capacity.notna()]

    # Filter so that only comparing on the same interpolation
    chosen_2 = chosen_2[(chosen_1.discharge_capacity.min() < chosen_2.discharge_capacity) &
                        (chosen_1.discharge_capacity.max() > chosen_2.discharge_capacity)]

    V = chosen_1.voltage.values
    Q = chosen_1.discharge_capacity.values

    # Threshold between values so that non-strictly monotonic values are removed
    # 1e7 roughly corresponds to the resolution of a 24 bit ADC, higher precision
    # would be unphysical
    d_capacity_min = (np.max(Q) - np.min(Q)) / 1e7
    if not np.all(np.diff(Q) >= -d_capacity_min):
        # Assuming that Q needs to be strictly increasing with threshold
        index_of_repeated = np.where(np.diff(Q) >= -d_capacity_min)[0]
        Q = np.delete(Q, index_of_repeated, axis=0)
        V = np.delete(V, index_of_repeated, axis=0)

    f = interp1d(Q, V, kind="cubic", fill_value="extrapolate", assume_sorted=False)

    v_2 = chosen_2.voltage.tolist()
    v_1 = f(chosen_2.discharge_capacity).tolist()
    v_diff = list_minus(v_1, v_2)

    if abs(np.var(v_diff)) > 1:
        print("weird voltage")
        return None
    else:
        result["var_v_diff"] = [np.var(v_diff)] 
        result["min_v_diff"] = [min(v_diff)]
        result["mean_v_diff"] = [np.mean(v_diff)]
        result["skew_v_diff"] = [skew(v_diff)]
        result["kurtosis_v_diff"] = [kurtosis(v_diff, fisher=False, bias=False)]
        result["sum_v_diff"] = [np.sum(np.absolute(v_diff))]
        result["sum_square_v_diff"] = [np.sum(np.square(v_diff))]

        return result


# TODO: this is a linear fit, we should use something
#  from a library, e.g. numpy.polyfit
# The equation I am using is based on the linear part of the curve 
def d_curve_fitting(x, y):
    """
    This function fits given data x and y into a linear function.

    Argument:
           relevant data x and y.

    Returns:
            the slope of the curve.
    """

    def test(x, a, b):
        return a * x + b

    param, param_cov = curve_fit(test, x, y)

    a = param[0]

    return a


def get_diffusion_coeff(processed_cycler_run, diag_pos, parameters_path=PROTOCOL_PARAMETERS_DIR):
    """
    This method calculates diffusion coefficients at different soc for a specified hppc cycle.
    (NOTE: The slope is proportional to 1/sqrt(D), and D here is interdiffusivity)

    Args:
        processed_cycler_run (beep.structure.ProcessedCyclerRun)
        diag_pos (int): diagnostic cycle occurrence for a specific <diagnostic_cycle_type>. e.g.
        if rpt_0.2C, occurs at cycle_index = [2, 37, 142, 244 ...], <diag_pos>=0 would correspond to cycle_index 2

    Returns:
        a dataframe with 8 entries, slope at different socs.
    """

    data = processed_cycler_run.diagnostic_data
    hppc_cycle = data.loc[data.cycle_type == "hppc"]
    cycles = hppc_cycle.cycle_index.unique()
    diag_num = cycles[diag_pos]

    selected_diag_df = hppc_cycle.loc[hppc_cycle.cycle_index == diag_num]
    selected_diag_df = selected_diag_df.sort_values(by="test_time")

    counters = []

    step_ind = get_step_index(processed_cycler_run,
                              cycle_type="hppc",
                              diag_pos=diag_pos,
                              parameters_path=parameters_path
                              )

    steps = [step_ind["hppc_long_rest"],
             step_ind["hppc_discharge_pulse"],
             step_ind["hppc_short_rest"],
             step_ind["hppc_charge_pulse"],
             step_ind["hppc_discharge_to_next_soc"]]

    for step in steps:
        counters.append(
            selected_diag_df[selected_diag_df.step_index == step].step_index_counter.unique().tolist()
        )

    result = pd.DataFrame()

    for i in range(1, min(len(counters[1]), len(counters[2]), 9)):
        discharge = selected_diag_df.loc[selected_diag_df.step_index_counter == counters[1][i]]
        rest = selected_diag_df.loc[selected_diag_df.step_index_counter == counters[2][i]]
        rest.loc[:, "diagnostic_time"] = rest["test_time"] - rest["test_time"].min()
        t_d = discharge.test_time.max() - discharge.test_time.min()
        v = rest.voltage
        t = rest.diagnostic_time
        x = np.sqrt(t + t_d) - np.sqrt(t)
        y = v - v.min()
        a = d_curve_fitting(
            x[round(2 * len(x) / 3): len(x)], y[round(2 * len(x) / 3): len(x)]
        )
        result["D_" + str(i)] = [a]

    return result


def get_diffusion_features(processed_cycler_run, diag_pos):
    """
    This method calculates changes in diffusion coefficient between a specified hppc cycle and the first one at
    different state of charge.

    Args:
        processed_cycler_run (beep.structure.ProcessedCyclerRun)
        diag_pos (int): diagnostic cycle occurence for a specific <diagnostic_cycle_type>. e.g.
        if rpt_0.2C, occurs at cycle_index = [2, 37, 142...], <diag_pos>=0 would correspond to cycle_index 2.

    Returns:
        a dataframe contains 8 normalized slope changes.

    """
    df_0 = get_diffusion_coeff(processed_cycler_run, 0)
    df = get_diffusion_coeff(processed_cycler_run, diag_pos)
    result = (df - df_0)/(df_0)
    return result


def get_fractional_quantity_remaining(
    processed_cycler_run, metric="discharge_energy", diagnostic_cycle_type="rpt_0.2C"
):
    """
    Determine relative loss of <metric> in diagnostic_cycles of type <diagnostic_cycle_type> after 100 regular cycles


    Args:
        processed_cycler_run (beep.structure.ProcessedCyclerRun): information about cycler run
        metric (str): column name to use for measuring degradation
        diagnostic_cycle_type (str): the diagnostic cycle to use for computing the amount of degradation

    Returns:
        a dataframe with cycle_index and corresponding degradation relative to the first measured value
    """
    summary_diag_cycle_type = processed_cycler_run.diagnostic_summary[
        (processed_cycler_run.diagnostic_summary.cycle_type == diagnostic_cycle_type)
        & (processed_cycler_run.diagnostic_summary.cycle_index > 100)
    ].reset_index()
    summary_diag_cycle_type = summary_diag_cycle_type[["cycle_index", metric]]
    summary_diag_cycle_type[metric] = (
        summary_diag_cycle_type[metric]
        / processed_cycler_run.diagnostic_summary[metric].iloc[0]
    )
    summary_diag_cycle_type.columns = ["cycle_index", "fractional_metric"]
    return summary_diag_cycle_type


def get_fractional_quantity_remaining_nx(
        processed_cycler_run,
        metric="discharge_energy",
        diagnostic_cycle_type="rpt_0.2C",
        parameters_path=PROTOCOL_PARAMETERS_DIR
):
    """
    Similar to get_fractional_quantity_remaining()
    Determine relative loss of <metric> in diagnostic_cycles of type <diagnostic_cycle_type>
    Also returns value of 'x', the discharge throughput passed by the first diagnostic
    and the value 'n' at each diagnostic

    Args:
        processed_cycler_run (beep.structure.ProcessedCyclerRun): information about cycler run
        metric (str): column name to use for measuring degradation
        diagnostic_cycle_type (str): the diagnostic cycle to use for computing the amount of degradation
        parameters_path (str): path to the parameters file for this run

    Returns:
        a dataframe with cycle_index, corresponding degradation relative to the first measured value, 'x',
        i.e. the discharge throughput passed by the first diagnostic
        and the value 'n' at each diagnostic, i.e. the equivalent scaling factor for lifetime using n*x
    """
    summary_diag_cycle_type = processed_cycler_run.diagnostic_summary[
        (processed_cycler_run.diagnostic_summary.cycle_type == diagnostic_cycle_type)
    ].reset_index()
    summary_diag_cycle_type = summary_diag_cycle_type[["cycle_index", "date_time_iso", metric]]

    # For the nx addition
    if 'energy' in metric:
        normalize_qty = 'discharge' + '_energy'
    else:
        normalize_qty = 'discharge' + '_capacity'

    normalize_qty_throughput = normalize_qty + '_throughput'
    regular_summary = processed_cycler_run.structured_summary.copy()
    regular_summary = regular_summary[regular_summary.cycle_index != 0]
    diagnostic_summary = processed_cycler_run.diagnostic_summary.copy()
    # TODO the line below should become superfluous
    regular_summary = regular_summary[
        ~regular_summary.cycle_index.isin(diagnostic_summary.cycle_index)]

    regular_summary.loc[:, normalize_qty_throughput] = regular_summary[normalize_qty].cumsum()
    diagnostic_summary.loc[:, normalize_qty_throughput] = diagnostic_summary[normalize_qty].cumsum()

    # Trim the cycle index in summary_diag_cycle_type to the max cycle in the regular cycles
    # (no partial cycles in the regular cycle summary) so that only full cycles are used for n
    summary_diag_cycle_type = summary_diag_cycle_type[summary_diag_cycle_type.cycle_index <=
                                                      regular_summary.cycle_index.max()]

    # Second gap in the regular cycles indicates the second set of diagnostics, bookending the
    # initial set of regular cycles.
    first_degradation_cycle = int(regular_summary.cycle_index[regular_summary.cycle_index.diff() > 1].iloc[0])
    last_initial_cycle = int(regular_summary.cycle_index[regular_summary.cycle_index <
                                                         first_degradation_cycle].iloc[-1])

    initial_regular_throughput = regular_summary[
            regular_summary.cycle_index == last_initial_cycle
        ][normalize_qty_throughput].values[0]

    summary_diag_cycle_type.loc[:, 'initial_regular_throughput'] = initial_regular_throughput

    summary_diag_cycle_type.loc[:, 'normalized_regular_throughput'] = summary_diag_cycle_type.apply(
        lambda x: (1 / initial_regular_throughput) *
        regular_summary[regular_summary.cycle_index < x['cycle_index']][normalize_qty_throughput].max(),
        axis=1
    )
    summary_diag_cycle_type['normalized_regular_throughput'].fillna(value=0, inplace=True)
    summary_diag_cycle_type.loc[:, 'normalized_diagnostic_throughput'] = summary_diag_cycle_type.apply(
        lambda x: (1 / initial_regular_throughput) *
        diagnostic_summary[diagnostic_summary.cycle_index < x['cycle_index']][normalize_qty_throughput].max(),
        axis=1
    )
    summary_diag_cycle_type['normalized_diagnostic_throughput'].fillna(value=0, inplace=True)
    # end of nx addition, calculate the fractional capacity compared to the first diagnostic cycle (reset)
    summary_diag_cycle_type.loc[:, metric] = (
        summary_diag_cycle_type[metric]
        / processed_cycler_run.diagnostic_summary[metric].iloc[0]
    )

    if "\\" in processed_cycler_run.metadata.protocol:
        protocol_name = processed_cycler_run.metadata.protocol.split("\\")[-1]
    else:
        _, protocol_name = os.path.split(processed_cycler_run.metadata.protocol)

    parameter_row, _ = parameters_lookup.get_protocol_parameters(protocol_name, parameters_path=parameters_path)

    summary_diag_cycle_type.loc[:, 'diagnostic_start_cycle'] = parameter_row['diagnostic_start_cycle'].values[0]
    summary_diag_cycle_type.loc[:, 'diagnostic_interval'] = parameter_row['diagnostic_interval'].values[0]

    # Calculate the epoch time stamp at each of the measurements for later comparison
    date_time_objs = pd.to_datetime(summary_diag_cycle_type["date_time_iso"])
    date_time_float = [
        calendar.timegm(t.timetuple()) if t is not pd.NaT else float("nan")
        for t in date_time_objs
    ]
    summary_diag_cycle_type.drop(columns=["date_time_iso"], inplace=True)
    summary_diag_cycle_type.loc[:, "epoch_time"] = date_time_float

    summary_diag_cycle_type.columns = ["cycle_index", "fractional_metric",
                                       "initial_regular_throughput", "normalized_regular_throughput",
                                       "normalized_diagnostic_throughput", "diagnostic_start_cycle",
                                       "diagnostic_interval", "epoch_time"]
    return summary_diag_cycle_type


def get_step_index(pcycler_run, cycle_type="hppc", diag_pos=0, parameters_path=PROTOCOL_PARAMETERS_DIR):
    """
        Gets the step indices of the diagnostic cycle which correspond to specific attributes

        Args:
            pcycler_run (beep.structure.ProcessedCyclerRun): processed data
            cycle_type (str): which diagnostic cycle type to evaluate
            diag_pos (int): which iteration of the diagnostic cycle to use (0 for first, 1 for second, -1 for last)

        Returns:
            dict: descriptive keys with step index as values
    """

    pulse_time = 120  # time in seconds used to decide if a current is a pulse or an soc change
    pulse_c_rate = 0.5  # c-rate to decide if a current is a discharge pulse
    rest_long_vs_short = 600  # time in seconds to decide if the rest is the long or short rest step
    soc_change_threshold = 0.05

    if "\\" in pcycler_run.metadata.protocol:
        protocol_name = pcycler_run.metadata.protocol.split("\\")[-1]
    else:
        _, protocol_name = os.path.split(pcycler_run.metadata.protocol)

    parameter_row, _ = parameters_lookup.get_protocol_parameters(protocol_name, parameters_path=parameters_path)

    step_indices_annotated = {}
    diag_data = pcycler_run.diagnostic_data
    cycles = diag_data.loc[diag_data.cycle_type == cycle_type]
    cycle = cycles[cycles.cycle_index == cycles.cycle_index.unique()[diag_pos]]

    if cycle_type == "hppc":
        for step in cycle.step_index.unique():
            cycle_step = cycle[(cycle.step_index == step)]
            median_crate = np.round(cycle_step.current.median() / parameter_row["capacity_nominal"].iloc[0], 2)
            mean_crate = np.round(cycle_step.current.mean() / parameter_row["capacity_nominal"].iloc[0], 2)
            remaining_time = cycle.test_time.max() - cycle_step.test_time.max()
            recurring = len(cycle_step.step_index_counter.unique()) > 1
            step_counter_duration = []
            for step_iter in cycle_step.step_index_counter.unique():
                cycle_step_iter = cycle_step[(cycle_step.step_index_counter == step_iter)]
                duration = cycle_step_iter.test_time.max() - cycle_step_iter.test_time.min()
                step_counter_duration.append(duration)
            median_duration = np.round(np.median(step_counter_duration), 0)

            if median_crate == 0.0:
                if median_duration > rest_long_vs_short:
                    step_indices_annotated["hppc_long_rest"] = step
                elif rest_long_vs_short >= median_duration > 0:
                    step_indices_annotated["hppc_short_rest"] = step
                else:
                    raise ValueError
            elif median_crate <= -pulse_c_rate and median_duration < pulse_time:
                step_indices_annotated["hppc_discharge_pulse"] = step
            elif median_crate >= pulse_c_rate and median_duration < pulse_time:
                step_indices_annotated["hppc_charge_pulse"] = step
            elif mean_crate != median_crate < 0 and remaining_time == 0.0 and not recurring:
                step_indices_annotated["hppc_final_discharge"] = step
            elif mean_crate == median_crate < 0 and remaining_time == 0.0 and not recurring:
                step_indices_annotated["hppc_final_discharge"] = step
            elif mean_crate == median_crate < 0 and abs(mean_crate * median_duration / 3600) > soc_change_threshold:
                step_indices_annotated["hppc_discharge_to_next_soc"] = step
            elif median_crate > 0 and median_duration > pulse_time:
                step_indices_annotated["hppc_charge_to_soc"] = step

    elif cycle_type == "rpt_0.2C" or cycle_type == "rpt_1C" or cycle_type == "rpt_2C" or cycle_type == "reset":
        for step in cycle.step_index.unique():
            cycle_step = cycle[(cycle.step_index == step)]
            median_crate = np.round(cycle_step.current.median() / parameter_row["capacity_nominal"].iloc[0], 2)
            if median_crate > 0:
                step_indices_annotated[cycle_type + "_charge"] = step
            elif median_crate < 0:
                step_indices_annotated[cycle_type + "_discharge"] = step
            else:
                raise ValueError
    else:
        raise NotImplementedError

    assert len(cycle.step_index.unique()) == len(step_indices_annotated.values())

    return step_indices_annotated


def check_diagnostic_validation(datapath):
    if not hasattr(datapath, "diagnostic_summary") & hasattr(
            datapath, "diagnostic_data"
    ):
        return False, "Datapath does not have diagnostic summary"
    if datapath.diagnostic_summary is None:
        return False, "Datapath does not have diagnostic summary"
    elif datapath.diagnostic_summary.empty:
        return False, "Datapath has empty diagnostic summary"
    else:
        return True, None



# Copyright [2020] [Toyota Research Institute]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import os
import unittest
import numpy as np
from sklearn.decomposition import PCA
from beep.features.principal_components import PrincipalComponents, pivot_data

from beep.tests.constants import TEST_FILE_DIR


class PrincipalComponentsTest(unittest.TestCase):
    def setUp(self):
        self.processed_run_path = os.path.join(
            TEST_FILE_DIR, "2017-06-30_2C-10per_6C_CH10_structure.json"
        )
        self.cycles_to_pca = np.linspace(20, 120, 20, dtype=int)
        self.cycles_to_test = np.linspace(121, 131, 6, dtype=int)
        json_obj = {"file_list": [self.processed_run_path], "run_list": [1]}
        json_string = json.dumps(json_obj)
        self.pc = PrincipalComponents.from_interpolated_data(
            json_string, cycles_to_pca=self.cycles_to_pca
        )

    def test_pivot_data(self):
        json_obj = {"file_list": [self.processed_run_path], "run_list": [1]}
        json_string = json.dumps(json_obj)
        df_to_pca = pivot_data(
            json_string, "discharge_capacity", "voltage", self.cycles_to_pca
        )
        self.assertEqual(df_to_pca.shape, (len(self.cycles_to_pca), 1000))

    def test_fit(self):
        self.assertIsInstance(self.pc.pca, PCA)
        self.assertEqual(self.pc.min_components, 4)

    def test_get_pca_embeddings(self):
        json_obj = {"file_list": [self.processed_run_path], "run_list": [1]}
        json_string = json.dumps(json_obj)
        df_to_pca = pivot_data(
            json_string, "discharge_capacity", "voltage", self.cycles_to_test
        )
        pca_embeddings = self.pc.get_pca_embeddings(df_to_pca)
        self.assertEqual(
            pca_embeddings.shape, (len(self.cycles_to_test), self.pc.n_components)
        )

    def test_get_pca_reconstruction(self):
        """
        Method to inverse transform PCA embeddings to reconstruct data
        """
        json_obj = {"file_list": [self.processed_run_path], "run_list": [1]}
        json_string = json.dumps(json_obj)
        df_to_pca = pivot_data(
            json_string, "discharge_capacity", "voltage", self.cycles_to_test
        )
        pca_embeddings = self.pc.get_pca_embeddings(df_to_pca)
        pca_reconstruction = self.pc.get_pca_reconstruction(pca_embeddings)
        self.assertEqual(pca_reconstruction.shape, (len(self.cycles_to_test), 1000))

    def test_get_reconstruction_errors(self):
        json_obj = {"file_list": [self.processed_run_path], "run_list": [1]}
        json_string = json.dumps(json_obj)
        df_to_pca = pivot_data(
            json_string, "discharge_capacity", "voltage", self.cycles_to_test
        )
        reconstruction_errors, outliers = self.pc.get_reconstruction_error_outliers(
            df_to_pca, threshold=1.5
        )
        self.assertAlmostEqual(reconstruction_errors[0], 0.0025532774, places=7)
        self.assertTrue(outliers[0])


if __name__ == "__main__":
    unittest.main()



# Copyright [2020] [Toyota Research Institute]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Unit tests related to Generating protocol files"""

import os
import unittest
import numpy as np
import datetime
import shutil
from copy import deepcopy

import difflib
import pandas as pd
from monty.tempfile import ScratchDir
from monty.serialization import dumpfn
from monty.os import makedirs_p


from beep import PROTOCOL_PARAMETERS_DIR
from beep.protocol import (
    PROCEDURE_TEMPLATE_DIR,
    SCHEDULE_TEMPLATE_DIR,
    BIOLOGIC_TEMPLATE_DIR,
)
from beep.protocol.generate_protocol import generate_protocol_files_from_csv, template_detection
from beep.utils.waveform import convert_velocity_to_power_waveform, RapidChargeWave
from beep.protocol.maccor import Procedure, \
    generate_maccor_waveform_file, insert_driving_parametersv1, insert_charging_parametersv1
from beep.protocol.arbin import Schedule
from beep.protocol.biologic import Settings
from beep.protocol.maccor_to_arbin import ProcedureToSchedule
from beep.utils import os_format, hash_file
from beep.tests.constants import TEST_FILE_DIR


this_dir = os.path.abspath(os.path.dirname(__file__))


class ProcedureTest(unittest.TestCase):
    def setUp(self):
        pass

    def test_convert_velocity_to_power_waveform(self):
        velocity_waveform_file = os.path.join(
            TEST_FILE_DIR, "LA4_velocity_waveform.txt"
        )
        df_velocity = pd.read_csv(velocity_waveform_file, sep="\t", header=0)
        df_power = convert_velocity_to_power_waveform(velocity_waveform_file, "mph")
        # Check input and output sizes
        self.assertEqual(len(df_velocity), len(df_power))
        self.assertTrue(any(df_power["power"] < 0))

    def test_generate_maccor_waveform_file_default(self):
        velocity_waveform_file = os.path.join(
            TEST_FILE_DIR, "LA4_velocity_waveform.txt"
        )
        with ScratchDir(".") as scratch_dir:

            df_power = convert_velocity_to_power_waveform(velocity_waveform_file, "mph")
            df_MWF = pd.read_csv(
                generate_maccor_waveform_file(
                    df_power, "test_LA4_waveform", scratch_dir
                ),
                sep="\t",
                header=None,
            )

            # Reference mwf file generated by the cycler for the same power waveform.
            df_MWF_ref = pd.read_csv(
                os.path.join(TEST_FILE_DIR, "LA4_ref_default.mwf"),
                sep="\t",
                header=None,
            )

            self.assertEqual(df_MWF.shape, df_MWF_ref.shape)

            # Check that the fourth column for charge/discharge limit is empty (default setting)
            self.assertTrue(df_MWF.iloc[:, 3].notnull().all())

            # Check that sum of durations equals length of the power timeseries
            self.assertEqual(df_MWF.iloc[:, 5].sum(), len(df_power))

            # Check that charge/discharge steps are identical
            self.assertTrue((df_MWF.iloc[:, 0] == df_MWF_ref.iloc[:, 0]).all())

            # Check that power values are close to each other (col 2)
            relative_differences = np.abs(
                (df_MWF.iloc[:, 2] - df_MWF_ref.iloc[:, 2]) / df_MWF_ref.iloc[:, 2]
            )
            self.assertLessEqual(
                np.mean(relative_differences) * 100, 0.01
            )  # mean percentage error < 0.01%

    def test_generate_maccor_waveform_file_custom(self):
        velocity_waveform_file = os.path.join(
            TEST_FILE_DIR, "US06_velocity_waveform.txt"
        )
        mwf_config = {
            "control_mode": "I",
            "value_scale": 1,
            "charge_limit_mode": "R",
            "charge_limit_value": 2,
            "discharge_limit_mode": "P",
            "discharge_limit_value": 3,
            "charge_end_mode": "V",
            "charge_end_operation": ">=",
            "charge_end_mode_value": 4.2,
            "discharge_end_mode": "V",
            "discharge_end_operation": "<=",
            "discharge_end_mode_value": 3,
            "report_mode": "T",
            "report_value": 10,
            "range": "A",
        }
        with ScratchDir(".") as scratch_dir:
            df_power = convert_velocity_to_power_waveform(velocity_waveform_file, "mph")
            df_MWF = pd.read_csv(
                generate_maccor_waveform_file(
                    df_power, "test_US06_waveform", scratch_dir, mwf_config=mwf_config
                ),
                sep="\t",
                header=None,
            )
            df_MWF_ref = pd.read_csv(
                os.path.join(TEST_FILE_DIR, "US06_reference_custom_settings.mwf"),
                sep="\t",
                header=None,
            )

            # Check dimensions with the reference mwf file
            self.assertEqual(df_MWF.shape, df_MWF_ref.shape)

            # Check that control_mode, charge/discharge state, limit and limit_value columns are identical.
            self.assertTrue(
                (df_MWF.iloc[:, [0, 1, 3, 4]] == df_MWF_ref.iloc[:, [0, 1, 3, 4]])
                .all()
                .all()
            )

            # Check that power values are close to each other (col 2)
            relative_differences = np.abs(
                (df_MWF.iloc[:, 2] - df_MWF_ref.iloc[:, 2]) / df_MWF_ref.iloc[:, 2]
            )
            self.assertLessEqual(
                np.mean(relative_differences) * 100, 0.01
            )  # mean percentage error < 0.01%

    def test_procedure_with_waveform(self):
        maccor_waveform_file = os.path.join(TEST_FILE_DIR, "LA4_8rep_lim.MWF")
        test_file = os.path.join(PROCEDURE_TEMPLATE_DIR, "diagnosticV2.000")
        procedure = Procedure.from_file(test_file)
        rest_step = procedure["MaccorTestProcedure"]["ProcSteps"]["TestStep"][2]
        end_step = procedure["MaccorTestProcedure"]["ProcSteps"]["TestStep"][-1]

        procedure["MaccorTestProcedure"]["ProcSteps"]["TestStep"] = procedure[
            "MaccorTestProcedure"
        ]["ProcSteps"]["TestStep"][:8]
        procedure["MaccorTestProcedure"]["ProcSteps"]["TestStep"][5:9] = deepcopy(
            procedure["MaccorTestProcedure"]["ProcSteps"]["TestStep"][4:8]
        )

        procedure.set("MaccorTestProcedure.ProcSteps.TestStep.5", deepcopy(rest_step))

        procedure.set("MaccorTestProcedure.ProcSteps.TestStep.9", deepcopy(rest_step))

        procedure.set("MaccorTestProcedure.ProcSteps.TestStep.10", deepcopy(end_step))

        procedure.set(
            "MaccorTestProcedure.ProcSteps.TestStep.5.Ends.EndEntry.0.Step", "007"
        )
        procedure.set(
            "MaccorTestProcedure.ProcSteps.TestStep.8.Ends.EndEntry.Step", "010"
        )
        procedure.set(
            "MaccorTestProcedure.ProcSteps.TestStep.9.Ends.EndEntry.0.Step", "011"
        )

        procedure = procedure.insert_maccor_waveform(6, maccor_waveform_file)
        for step in procedure["MaccorTestProcedure"]["ProcSteps"]["TestStep"]:
            print(step["Ends"])
            if step["StepType"] in ["Charge", "Dischrge", "Rest"]:
                step["Ends"]["EndEntry"][-1]["Step"] = "011"
                step["Ends"]["EndEntry"][-2]["Step"] = "011"

        steps = [
            x["StepType"]
            for x in procedure["MaccorTestProcedure"]["ProcSteps"]["TestStep"]
        ]
        self.assertEqual(
            steps,
            [
                "Rest",
                "Charge",
                "Rest",
                "Do 1",
                "Charge",
                "Rest",
                "FastWave",
                "AdvCycle",
                "Loop 1",
                "Rest",
                "End",
            ],
        )
        self.assertEqual(
            procedure["MaccorTestProcedure"]["ProcSteps"]["TestStep"][5]["Ends"][
                "EndEntry"
            ][0]["Step"],
            "007",
        )
        self.assertEqual(
            procedure["MaccorTestProcedure"]["ProcSteps"]["TestStep"][6]["StepType"],
            "FastWave",
        )

        with ScratchDir(".") as scratch_dir:
            local_name = "test_mwf_LA4_lim.000"
            procedure.to_file(os.path.join(scratch_dir, local_name))
            # Uncomment line below to keep the output in the test file directory
            # shutil.copyfile(os.path.join(scratch_dir, local_name), os.path.join(TEST_FILE_DIR, local_name))

    def test_generate_maccor_waveform_from_power(self):
        power_waveform_file = os.path.join(TEST_FILE_DIR, "LA4_power_profile.csv")
        mwf_config = {
            "control_mode": "P",
            "value_scale": 30,
            "charge_limit_mode": "V",
            "charge_limit_value": 4.2,
            "discharge_limit_mode": "V",
            "discharge_limit_value": 2.7,
            "charge_end_mode": "V",
            "charge_end_operation": ">=",
            "charge_end_mode_value": 4.25,
            "discharge_end_mode": "V",
            "discharge_end_operation": "<=",
            "discharge_end_mode_value": 2.5,
            "report_mode": "T",
            "report_value": 3.0000,
            "range": "A",
        }
        with ScratchDir(".") as scratch_dir:
            waveform_name = "LA4_8rep_lim"
            df_power = pd.read_csv(power_waveform_file)
            df_power = pd.concat(
                [
                    df_power,
                    df_power,
                    df_power,
                    df_power,
                    df_power,
                    df_power,
                    df_power,
                    df_power,
                ]
            )
            df_power.drop(columns=["power"], inplace=True)
            df_power.rename(columns={"power_scaled": "power"}, inplace=True)
            df_MWF = pd.read_csv(
                generate_maccor_waveform_file(
                    df_power, waveform_name, scratch_dir, mwf_config=mwf_config
                ),
                sep="\t",
                header=None,
            )
            sign = df_MWF[0].apply(lambda x: -1 if x is "D" else 1)
            energy = df_MWF[2] * df_MWF[5] * sign
            self.assertLess(df_MWF[5].sum(), 10961)
            self.assertLess(energy.sum(), -20000)
            self.assertGreater(40, df_MWF[2].max())
            self.assertLess(-40, df_MWF[2].min())
            # Uncomment line below to keep the output in the test file directory
            shutil.copyfile(
                os.path.join(scratch_dir, waveform_name + ".MWF"),
                os.path.join(TEST_FILE_DIR, waveform_name + ".MWF"),
            )

    def test_generate_maccor_waveform_from_current(self):
        charging_c_rates = [0.5, 2.5, 2.0, 0.2]
        soc_points = [0.05, 0.25, 0.65, 0.8]
        final_c_rate = charging_c_rates[-1]
        soc_initial = soc_points[0]
        soc_final = soc_points[-1]
        max_c_rate = 3.0
        min_c_rate = 0.2
        nominal_capacity = 4.84

        mwf_config = {
            "control_mode": "I",
            "value_scale": 1,
            "charge_limit_mode": "V",
            "charge_limit_value": 4.2,
            "discharge_limit_mode": "V",
            "discharge_limit_value": 2.7,
            "charge_end_mode": "V",
            "charge_end_operation": ">=",
            "charge_end_mode_value": 4.25,
            "discharge_end_mode": "V",
            "discharge_end_operation": "<=",
            "discharge_end_mode_value": 2.5,
            "report_mode": "T",
            "report_value": 3.0000,
            "range": "A",
        }
        with ScratchDir(".") as scratch_dir:
            waveform_name = "smooth_current_test"
            charging = RapidChargeWave(final_c_rate, soc_initial, soc_final, max_c_rate, min_c_rate)
            current_smooth, current_multi, time = charging.get_currents_with_uniform_time_basis(charging_c_rates,
                                                                                                soc_points)
            df_smooth = pd.DataFrame({"current": current_smooth, "time": time})
            mwf_config["value_scale"] = nominal_capacity * max(df_smooth["current"])

            generate_maccor_waveform_file(
                df_smooth,
                waveform_name,
                scratch_dir,
                mwf_config=mwf_config
            )

            df_mwf = pd.read_csv(
                os.path.join(scratch_dir, waveform_name + '.MWF'),
                sep="\t",
                header=None,
            )
            self.assertTrue(np.all(df_mwf.iloc[:, [1, ]] == 'I'))
            self.assertGreater(df_mwf.loc[:, 2].max(), nominal_capacity * max(charging_c_rates))
            self.assertEqual(len(df_mwf), 1572)
            # shutil.copyfile(os.path.join(scratch_dir, waveform_name + '.MWF'),
            #                 os.path.join(TEST_FILE_DIR, waveform_name + '.MWF'))

            waveform_name = "multistep_current_test"
            charging = RapidChargeWave(final_c_rate, soc_initial, soc_final, max_c_rate, min_c_rate)
            current_smooth, current_multi, time = charging.get_currents_with_uniform_time_basis(charging_c_rates,
                                                                                                soc_points)
            df_smooth = pd.DataFrame({"current": current_multi, "time": time})
            mwf_config["value_scale"] = nominal_capacity * max(df_smooth["current"])

            generate_maccor_waveform_file(
                df_smooth,
                waveform_name,
                scratch_dir,
                mwf_config=mwf_config
            )
            df_mwf = pd.read_csv(
                os.path.join(scratch_dir, waveform_name + '.MWF'),
                sep="\t",
                header=None,
            )
            self.assertTrue(np.all(df_mwf.iloc[:, [1, ]] == 'I'))
            self.assertGreaterEqual(df_mwf.loc[:, 2].max(), nominal_capacity * max(charging_c_rates))
            self.assertEqual(len(df_mwf), 15)
            # Uncomment line below to keep the output in the test file directory
            # shutil.copyfile(os.path.join(scratch_dir, waveform_name + '.MWF'),
            #                 os.path.join(TEST_FILE_DIR, waveform_name + '.MWF'))


class GenerateProcedureTest(unittest.TestCase):
    def setUp(self):
        pass

    def test_io(self):
        test_file = os.path.join(TEST_FILE_DIR, "xTESLADIAG_000003_CH68.000")
        json_file = os.path.join(TEST_FILE_DIR, "xTESLADIAG_000003_CH68.json")
        test_out = "test1.000"

        procedure = Procedure.from_file(os.path.join(TEST_FILE_DIR, test_file))
        with ScratchDir("."):
            dumpfn(procedure, json_file)
            procedure.to_file(test_out)
            hash1 = hash_file(test_file)
            hash2 = hash_file(test_out)
            if hash1 != hash2:
                original = open(test_file).readlines()
                parsed = open(test_out).readlines()
                self.assertFalse(list(difflib.unified_diff(original, parsed)))
                for line in difflib.unified_diff(original, parsed):
                    self.assertIsNotNone(line)

        test_file = os.path.join(TEST_FILE_DIR, "xTESLADIAG_000004_CH69.000")
        json_file = os.path.join(TEST_FILE_DIR, "xTESLADIAG_000004_CH69.json")
        test_out = "test2.000"

        procedure = Procedure.from_file(os.path.join(TEST_FILE_DIR, test_file))
        with ScratchDir("."):
            dumpfn(procedure, json_file)
            procedure.to_file(test_out)
            hash1 = hash_file(test_file)
            hash2 = hash_file(test_out)
            if hash1 != hash2:
                original = open(test_file).readlines()
                parsed = open(test_out).readlines()
                self.assertFalse(list(difflib.unified_diff(original, parsed)))
                for line in difflib.unified_diff(original, parsed):
                    self.assertIsNotNone(line)

    def test_generate_proc_exp(self):
        test_file = os.path.join(TEST_FILE_DIR, "EXP.000")
        json_file = os.path.join(TEST_FILE_DIR, "EXP.json")
        test_out = "test_EXP.000"
        test_parameters = ["4.2", "2.0C", "2.0C"]
        procedure = Procedure.from_exp(*test_parameters)
        with ScratchDir("."):
            dumpfn(procedure, json_file)
            procedure.to_file(test_out)
            hash1 = hash_file(test_file)
            hash2 = hash_file(test_out)
            if hash1 != hash2:
                original = open(test_file).readlines()
                parsed = open(test_out).readlines()
                self.assertFalse(list(difflib.unified_diff(original, parsed)))
                for line in difflib.unified_diff(original, parsed):
                    self.assertIsNotNone(line)

    def test_missing(self):
        test_parameters = ["EXP", "4.2", "2.0C", "2.0C"]
        template = os.path.join(TEST_FILE_DIR, "EXP_missing.000")
        self.assertRaises(
            UnboundLocalError, Procedure.from_exp, *test_parameters[1:] + [template]
        )

    def test_prediag_with_waveform(self):
        maccor_waveform_file = os.path.join(TEST_FILE_DIR, "LA4_8rep_lim.MWF")
        test_file = os.path.join(PROCEDURE_TEMPLATE_DIR, "diagnosticV3.000")
        csv_file = os.path.join(TEST_FILE_DIR, "PredictionDiagnostics_parameters.csv")
        protocol_params_df = pd.read_csv(csv_file)
        index = 1
        protocol_params_df.iloc[
            index, protocol_params_df.columns.get_loc("capacity_nominal")
        ] = 3.71
        protocol_params = protocol_params_df.iloc[index]
        diag_params_df = pd.read_csv(
            os.path.join(PROCEDURE_TEMPLATE_DIR, "PreDiag_parameters - DP.csv")
        )
        diagnostic_params = diag_params_df[
            diag_params_df["diagnostic_parameter_set"]
            == protocol_params["diagnostic_parameter_set"]
        ].squeeze()

        procedure = Procedure.generate_procedure_regcyclev3(index, protocol_params)
        procedure.generate_procedure_diagcyclev3(
            protocol_params["capacity_nominal"], diagnostic_params
        )

        steps = [
            x["StepType"]
            for x in procedure["MaccorTestProcedure"]["ProcSteps"]["TestStep"]
        ]
        print(steps)
        start = 27
        reg_cycle_steps = [
            "Do 1",
            "Charge",
            "Charge",
            "Charge",
            "Rest",
            "Dischrge",
            "Rest",
            "AdvCycle",
            "Loop 1",
        ]
        reg_steps_len = len(reg_cycle_steps)
        self.assertEqual(steps[start : start + reg_steps_len], reg_cycle_steps)
        start = 59
        reg_cycle_steps = [
            "Do 2",
            "Charge",
            "Charge",
            "Charge",
            "Rest",
            "Dischrge",
            "Rest",
            "AdvCycle",
            "Loop 2",
        ]
        reg_steps_len = len(reg_cycle_steps)
        self.assertEqual(steps[start: start + reg_steps_len], reg_cycle_steps)
        print(procedure["MaccorTestProcedure"]["ProcSteps"]["TestStep"][32])
        print(procedure["MaccorTestProcedure"]["ProcSteps"]["TestStep"][64])

        procedure.insert_maccor_waveform(32, maccor_waveform_file)
        procedure.insert_maccor_waveform(64, maccor_waveform_file)
        self.assertEqual(
            procedure["MaccorTestProcedure"]["ProcSteps"]["TestStep"][32]["StepType"],
            "FastWave",
        )
        self.assertEqual(
            procedure["MaccorTestProcedure"]["ProcSteps"]["TestStep"][64]["StepType"],
            "FastWave",
        )
        with ScratchDir(".") as scratch_dir:
            driving_test_name = "Drive_test20200716.000"
            procedure.to_file(driving_test_name)
            # Uncomment line below to keep the output in the test file directory
            # shutil.copyfile(os.path.join(scratch_dir, driving_test_name),
            #                 os.path.join(TEST_FILE_DIR, driving_test_name))

    def test_waveform_from_csv(self):
        csv_file = os.path.join(TEST_FILE_DIR,
                                "data-share",
                                "raw",
                                "parameters",
                                "Drive_parameters - GP.csv")
        protocol_params_df = pd.read_csv(csv_file)
        # Max power in watts in the profiles (should be less than max power limit on cyclers)
        MAX_PROFILE_CURRENT = 15
        MIN_PROFILE_VOLTAGE = 2.7
        MAX_PROFILE_POWER = MAX_PROFILE_CURRENT * MIN_PROFILE_VOLTAGE

        new_files = []
        names = []
        waveform_names = []
        result = ""
        message = {"comment": "", "error": ""}
        with ScratchDir(".") as scratch_dir:
            output_directory = scratch_dir
            os.makedirs(os.path.join(output_directory, "procedures"))
            for index, protocol_params in protocol_params_df.iterrows():
                template = protocol_params["template"]
                filename_prefix = "_".join(
                    [
                        protocol_params["project_name"],
                        "{:06d}".format(protocol_params["seq_num"]),
                    ]
                )
                if template == "diagnosticV5.000":
                    diag_params_df = pd.read_csv(
                        os.path.join(PROCEDURE_TEMPLATE_DIR, "PreDiag_parameters - DP.csv")
                    )
                    diagnostic_params = diag_params_df[
                        diag_params_df["diagnostic_parameter_set"]
                        == protocol_params["diagnostic_parameter_set"]
                        ].squeeze()
                    mwf_dir = os.path.join(scratch_dir, "mwf_files")
                    waveform_name = insert_driving_parametersv1(protocol_params,
                                                                waveform_directory=mwf_dir)
                    template_fullpath = os.path.join(PROCEDURE_TEMPLATE_DIR, template)
                    protocol = Procedure.generate_procedure_drivingv1(index,
                                                                      protocol_params,
                                                                      waveform_name,
                                                                      template=template_fullpath)
                    protocol.generate_procedure_diagcyclev3(
                        protocol_params["capacity_nominal"], diagnostic_params
                    )
                    filename = "{}.000".format(filename_prefix)
                    filename = os.path.join(output_directory, "procedures", filename)
                    waveform_names.append(os.path.split(waveform_name)[-1])

                if not os.path.isfile(filename):
                    protocol.to_file(filename)
                    new_files.append(filename)
                    names.append(filename_prefix + "_")

                waveform_df = pd.read_csv(waveform_name, sep="\t", header=None)
                power_df = waveform_df[waveform_df[1] == "P"]
                self.assertLessEqual(power_df[2].abs().max(), MAX_PROFILE_POWER)
                self.assertGreaterEqual(power_df[2].abs().max(), 0)

                current_df = waveform_df[waveform_df[1] == "I"]
                self.assertLessEqual(current_df[2].abs().max(), MAX_PROFILE_CURRENT)
                self.assertGreaterEqual(current_df[2].abs().max(), 0)

                discharge_df = waveform_df[waveform_df[0] == "D"]
                self.assertGreaterEqual(discharge_df[4].abs().max(), MIN_PROFILE_VOLTAGE)

                self.assertEqual(protocol.get_path("MaccorTestProcedure.ProcSteps.TestStep.32.StepType"), 'FastWave')
                self.assertEqual(protocol.get_path("MaccorTestProcedure.ProcSteps.TestStep.64.StepType"), 'FastWave')

#                Uncomment to test waveform names
#                wave_value = os.path.split(waveform_name)[-1].split(".")[0]
#                self.assertEqual(protocol.get_path("MaccorTestProcedure.ProcSteps.TestStep.32.StepValue"),
#                                 wave_value)
#                self.assertEqual(protocol.get_path("MaccorTestProcedure.ProcSteps.TestStep.64.StepValue"),
#                                 wave_value)

            self.assertEqual(len(os.listdir(os.path.join(output_directory, "procedures"))), 36)
            self.assertEqual(len(os.listdir(os.path.join(output_directory, "mwf_files"))), 18)

        test_names = ['US06_x4_24W.MWF', 'LA4_x4_10W.MWF',
                      'US06_x4_32W.MWF', 'LA4_x4_14W.MWF',
                      'US06_x4_40W.MWF', 'LA4_x4_18W.MWF',
                      'US06_x8_24W.MWF', 'LA4_x8_10W.MWF',
                      'US06_x8_32W.MWF', 'LA4_x8_14W.MWF',
                      'US06_x8_40W.MWF', 'LA4_x8_18W.MWF',
                      'US06_x12_24W.MWF', 'LA4_x12_10W.MWF',
                      'US06_x12_32W.MWF', 'LA4_x12_14W.MWF',
                      'US06_x12_40W.MWF', 'LA4_x12_18W.MWF',
                      'US06_x4_24W.MWF', 'LA4_x4_10W.MWF',
                      'US06_x4_32W.MWF', 'LA4_x4_14W.MWF',
                      'US06_x4_40W.MWF', 'LA4_x4_18W.MWF',
                      'US06_x8_24W.MWF', 'LA4_x8_10W.MWF',
                      'US06_x8_32W.MWF', 'LA4_x8_14W.MWF',
                      'US06_x8_40W.MWF', 'LA4_x8_18W.MWF',
                      'US06_x12_24W.MWF', 'LA4_x12_10W.MWF',
                      'US06_x12_32W.MWF', 'LA4_x12_14W.MWF',
                      'US06_x12_40W.MWF', 'LA4_x12_18W.MWF']
        self.assertListEqual(test_names, waveform_names)

    def test_from_csv(self):
        csv_file = os.path.join(TEST_FILE_DIR, "parameter_test.csv")

        # Test basic functionality
        with ScratchDir(".") as scratch_dir:
            makedirs_p(os.path.join(scratch_dir, "procedures"))
            makedirs_p(os.path.join(scratch_dir, "names"))
            _new_files, _generation_failures, result, message = generate_protocol_files_from_csv(
                csv_file, output_directory=scratch_dir
            )
            self.assertEqual(
                len(os.listdir(os.path.join(scratch_dir, "procedures"))), 3
            )
            self.assertEqual(result, "error")

        # Test overwriting file functionality
        with ScratchDir(".") as scratch_dir:
            makedirs_p(os.path.join(scratch_dir, "procedures"))
            makedirs_p(os.path.join(scratch_dir, "names"))
            dumpfn({"hello": "world"}, os.path.join("procedures", "name_000007.000"))
            _new_files, generation_failures, result, message = generate_protocol_files_from_csv(
                csv_file, output_directory=scratch_dir
            )
            # To test avoiding overwriting file uncomment lines below
            # post_file = loadfn(os.path.join("procedures", "name_000007.000"))
            # self.assertEqual(post_file, {"hello": "world"})
            self.assertEqual(
                len(os.listdir(os.path.join(scratch_dir, "procedures"))), 3
            )
            self.assertEqual(result, "error")
            self.assertEqual(
                message,
                {
                    "comment": "Generated 3 of 33 protocols",
                    "error": "Failed to generate 30 of 33 protocols",
                },
            )
            self.assertEqual(
                generation_failures[0],
                {
                    "comment": "Unable to find template: EXP-SOH.000",
                    "error": "Not Found"
                },
            )

    def test_from_csv_2(self):
        csv_file = os.path.join(TEST_FILE_DIR, "PredictionDiagnostics_parameters.csv")

        # Test basic functionality
        with ScratchDir(".") as scratch_dir:
            makedirs_p(os.path.join(scratch_dir, "procedures"))
            makedirs_p(os.path.join(scratch_dir, "names"))
            _new_files, _generation_failures, result, message = generate_protocol_files_from_csv(
                csv_file, output_directory=scratch_dir
            )
            self.assertEqual(result, "success")
            self.assertEqual(message, {"comment": "Generated 2 of 2 protocols", "error": ""})
            self.assertEqual(
                len(os.listdir(os.path.join(scratch_dir, "procedures"))), 2
            )

            original = open(
                os.path.join(PROCEDURE_TEMPLATE_DIR, "diagnosticV2.000")
            ).readlines()
            parsed = open(
                os.path.join(
                    os.path.join(scratch_dir, "procedures"),
                    "PredictionDiagnostics_000000.000",
                )
            ).readlines()
            self.assertFalse(list(difflib.unified_diff(original, parsed)))
            for line in difflib.unified_diff(original, parsed):
                self.assertIsNotNone(line)

            original = open(
                os.path.join(PROCEDURE_TEMPLATE_DIR, "diagnosticV3.000")
            ).readlines()
            parsed = open(
                os.path.join(
                    os.path.join(scratch_dir, "procedures"),
                    "PredictionDiagnostics_000196.000",
                )
            ).readlines()
            diff = list(difflib.unified_diff(original, parsed))
            diff_expected = [
                "--- \n",
                "+++ \n",
                "@@ -27,7 +27,7 @@\n",
                "           <SpecialType> </SpecialType>\n",
                "           <Oper> = </Oper>\n",
                "           <Step>002</Step>\n",
                "-          <Value>03:00:00</Value>\n",
                "+          <Value>03:12:00</Value>\n",
                "         </EndEntry>\n",
                "         <EndEntry>\n",
                "           <EndType>Voltage </EndType>\n",
            ]
            self.assertEqual(diff, diff_expected)
            for line in difflib.unified_diff(original, parsed):
                self.assertIsNotNone(line)

            _, namefile = os.path.split(csv_file)
            namefile = namefile.split("_")[0] + "_names_"
            namefile = (
                namefile + datetime.datetime.now().strftime("%Y%m%d_%H%M") + ".csv"
            )
            names_test = open(os.path.join(scratch_dir, "names", namefile)).readlines()
            self.assertEqual(
                names_test,
                ["PredictionDiagnostics_000000_\n", "PredictionDiagnostics_000196_\n"],
            )

    @unittest.skip
    def test_from_csv_3(self):

        csv_file_list = os.path.join(TEST_FILE_DIR, "PreDiag_parameters - GP.csv")
        makedirs_p(os.path.join(TEST_FILE_DIR, "procedures"))
        makedirs_p(os.path.join(TEST_FILE_DIR, "names"))
        generate_protocol_files_from_csv(csv_file_list, output_directory=TEST_FILE_DIR)
        if os.path.isfile(os.path.join(TEST_FILE_DIR, "procedures", ".DS_Store")):
            os.remove(os.path.join(TEST_FILE_DIR, "procedures", ".DS_Store"))
        self.assertEqual(
            len(os.listdir(os.path.join(TEST_FILE_DIR, "procedures"))), 386
        )

    def test_charging_waveform_from_csv(self):
        csv_file = os.path.join(TEST_FILE_DIR,
                                "data-share",
                                "raw",
                                "parameters",
                                "RapidC_parameters - GP.csv")
        protocol_params_df = pd.read_csv(csv_file)

        with ScratchDir(".") as scratch_dir:
            reg_params = protocol_params_df.iloc[1]
            output_directory = scratch_dir

            file_name = insert_charging_parametersv1(reg_params,
                                                     waveform_directory=output_directory,
                                                     max_c_rate=3.0, min_c_rate=0.2)
            self.assertTrue(os.path.isfile(os.path.join(output_directory, file_name)))
            self.assertEqual(os.path.split(file_name)[1], "RapidC_smooth_101.MWF")

    def test_charging_protocol_with_waveform(self):
        csv_file = os.path.join(TEST_FILE_DIR,
                                "data-share",
                                "raw",
                                "parameters",
                                "RapidC_parameters - GP.csv")
        protocol_params_df = pd.read_csv(csv_file)

        successfully_generated_files = []
        file_generation_failures = []
        names = []
        result = ""
        message = {"comment": "", "error": ""}
        with ScratchDir(".") as scratch_dir:
            output_directory = scratch_dir
            os.makedirs(os.path.join(output_directory, "procedures"))
            for index, protocol_params in protocol_params_df.iterrows():
                template = protocol_params["template"]
                filename_prefix = "_".join(
                    [
                        protocol_params["project_name"],
                        "{:06d}".format(protocol_params["seq_num"]),
                    ]
                )
                if template in ["diagnosticV3.000", "diagnosticV4.000"]:
                    diag_params_df = pd.read_csv(
                        os.path.join(PROCEDURE_TEMPLATE_DIR, "PreDiag_parameters - DP.csv")
                    )
                    diagnostic_params = diag_params_df[
                        diag_params_df["diagnostic_parameter_set"]
                        == protocol_params["diagnostic_parameter_set"]
                        ].squeeze()
                    template_fullpath = os.path.join(PROCEDURE_TEMPLATE_DIR, template)
                    if protocol_params["project_name"] == "RapidC":
                        mwf_dir = os.path.join(output_directory, "mwf_files")
                        waveform_name = insert_charging_parametersv1(protocol_params,
                                                                     waveform_directory=mwf_dir)
                        protocol = Procedure.generate_procedure_chargingv1(index,
                                                                           protocol_params,
                                                                           waveform_name,
                                                                           template=template_fullpath)
                    else:
                        print("Test did not recognize project name correctly")
                        raise NotImplementedError

                    protocol.generate_procedure_diagcyclev3(
                        protocol_params["capacity_nominal"], diagnostic_params
                    )
                    filename = "{}.000".format(filename_prefix)
                    filename = os.path.join(output_directory, "procedures", filename)

                if not os.path.isfile(filename):
                    protocol.to_file(filename)
                    successfully_generated_files.append(filename)
                    names.append(filename_prefix + "_")

                # Reference mwf file generated by the cycler for the same power waveform.
                df_mwf = pd.read_csv(
                    waveform_name,
                    sep="\t",
                    header=None,
                )
                self.assertTrue(np.all(df_mwf.iloc[:, [1, ]] == 'I'))
                self.assertEqual(df_mwf.loc[:, 4].max(), protocol_params["charge_cutoff_voltage"] - 0.01)
                self.assertEqual(df_mwf.loc[:, 4].min(), protocol_params["charge_cutoff_voltage"] - 0.01)

            self.assertTrue(os.path.isfile(filename))
            self.assertEqual(len(os.listdir(mwf_dir)), 60)

#            Uncomment the following tests if not running on Windows
#            parsed = open(
#                os.path.join(scratch_dir, "procedures", "RapidC_000100.000")
#            )

#            lines = parsed.readlines()
#            self.assertEqual(lines[972], "      <StepType>FastWave</StepType>\n")
#            self.assertEqual(lines[974], "      <StepValue>RapidC_smooth_100</StepValue>\n")
#            self.assertEqual(lines[2033], "      <StepType>FastWave</StepType>\n")
#            self.assertEqual(lines[2035], "      <StepValue>RapidC_smooth_100</StepValue>\n")

#            parsed.close()


class GenerateProtocolTest(unittest.TestCase):
    def setup(self):
        pass

    # code for lean folder (not on Windows)

    def tearDown(self) -> None:
        for d in ("mwf_files", "names", "procedures", "settings"):
            p = os.path.join(this_dir, d)
            if os.path.exists(p):
                shutil.rmtree(p)

    def test_template_detection(self):
        length_1 = template_detection(os.path.join(PROCEDURE_TEMPLATE_DIR, "diagnosticV1.000"))
        self.assertEqual(72, length_1)

        length_2 = template_detection(os.path.join(PROCEDURE_TEMPLATE_DIR, "diagnosticV5.000"))
        self.assertEqual(96, length_2)

        length_3 = template_detection(os.path.join(PROCEDURE_TEMPLATE_DIR, "EXP.000"))
        self.assertEqual(23, length_3)

    def test_multiple_protocol_generation(self):

        for csv, res in {
            "Form_parameters - GP.csv": 16,
            "Drive_parameters - GP.csv": 36,
            "RapidC_parameters - GP.csv": 60,
            # "unsupported_test.csv": 0
        }.items():

            csv_file = os.path.join(PROTOCOL_PARAMETERS_DIR, csv)
            results = generate_protocol_files_from_csv(csv_file, output_directory=this_dir)
            output_files, file_generation_failures, result, message = results
            self.assertEqual(len(output_files), res)
            self.assertEqual(len(file_generation_failures), 0)


class ProcedureToScheduleTest(unittest.TestCase):
    def setUp(self):
        pass

    def test_single_step_conversion(self):
        procedure = Procedure()

        templates = PROCEDURE_TEMPLATE_DIR

        test_file = "diagnosticV3.000"
        json_file = "test.json"

        proc_dict = procedure.from_file(os.path.join(templates, test_file))
        test_step_dict = proc_dict["MaccorTestProcedure"]["ProcSteps"]["TestStep"]

        converter = ProcedureToSchedule(test_step_dict)
        step_index = 5
        step_name_list, step_flow_ctrl = converter.create_metadata()

        self.assertEqual(step_flow_ctrl[7], "5-reset cycle C/20")
        self.assertEqual(step_flow_ctrl[68], "38-reset cycle")

        step_arbin = converter.compile_to_arbin(
            test_step_dict[step_index], step_index, step_name_list, step_flow_ctrl
        )
        self.assertEqual(step_arbin["m_szLabel"], "6-None")
        self.assertEqual(step_arbin["Limit0"]["m_szGotoStep"], "Next Step")
        self.assertEqual(step_arbin["Limit0"]["Equation0_szLeft"], "PV_CHAN_Voltage")
        self.assertEqual(
            step_arbin["Limit2"]["m_szGotoStep"], "70-These are the 2 reset cycles"
        )

        step_index = 8
        step_arbin = converter.compile_to_arbin(
            test_step_dict[step_index], step_index, step_name_list, step_flow_ctrl
        )

        self.assertEqual(
            step_arbin["Limit0"]["Equation0_szLeft"], "PV_CHAN_CV_Stage_Current"
        )
        self.assertEqual(
            step_arbin["Limit0"]["Equation0_szRight"],
            test_step_dict[step_index]["Ends"]["EndEntry"][0]["Value"],
        )

    def test_serial_conversion(self):
        procedure = Procedure()

        templates = PROCEDURE_TEMPLATE_DIR

        test_file = "diagnosticV3.000"
        json_file = "test.json"

        proc_dict = procedure.from_file(os.path.join(templates, test_file))

        test_step_dict = proc_dict["MaccorTestProcedure"]["ProcSteps"]["TestStep"]

        converter = ProcedureToSchedule(test_step_dict)
        step_name_list, step_flow_ctrl = converter.create_metadata()

        for step_index, step in enumerate(test_step_dict):
            if "Loop" in step["StepType"]:
                print(step_index, step)
            step_arbin = converter.compile_to_arbin(
                test_step_dict[step_index], step_index, step_name_list, step_flow_ctrl
            )
            if "Loop" in step["StepType"]:
                self.assertEqual(step_arbin["m_szStepCtrlType"], "Set Variable(s)")
                self.assertEqual(step_arbin["m_uLimitNum"], "2")
            if step_index == 15:
                self.assertEqual(step_arbin["Limit0"]["m_szGotoStep"], "11-None")
                self.assertEqual(step_arbin["Limit1"]["m_szGotoStep"], "Next Step")

    def test_schedule_creation(self):
        protocol_params_dict = {
         'project_name': ['PreDiag'],
         'seq_num': [100],
         'template': ['diagnosticV3.000'],
         'charge_constant_current_1': [1],
         'charge_percent_limit_1': [30],
         'charge_constant_current_2': [1],
         'charge_cutoff_voltage': [3.6],
         'charge_constant_voltage_time': [30],
         'charge_rest_time': [5],
         'discharge_constant_current': [1],
         'discharge_cutoff_voltage': [3.0],
         'discharge_rest_time': [15],
         'cell_temperature_nominal': [25],
         'cell_type': ['Tesla_Model3_21700'],
         'capacity_nominal': [1.1],
         'diagnostic_type': ['HPPC+RPT'],
         'diagnostic_parameter_set': ['Tesla21700'],
         'diagnostic_start_cycle': [30],
         'diagnostic_interval': [100]
         }
        procedure_to_convert = 'test_procedure.000'
        with ScratchDir('.') as scratch_dir:
            protocol_params_df = pd.DataFrame.from_dict(protocol_params_dict)
            protocol_params = protocol_params_df.iloc[[0]].squeeze()

            diag_params_df = pd.read_csv(os.path.join(PROCEDURE_TEMPLATE_DIR,
                                                      "PreDiag_parameters - DP.csv"))
            diagnostic_params = diag_params_df[diag_params_df['diagnostic_parameter_set'] == 'A123LFP']

            procedure = Procedure.generate_procedure_regcyclev3(0, protocol_params)
            procedure.generate_procedure_diagcyclev3(
                protocol_params["capacity_nominal"], diagnostic_params
            )
            procedure.set_skip_to_end_diagnostic(3.8, 2.0, step_key='070')
            self.assertEqual(procedure['MaccorTestProcedure']['ProcSteps']['TestStep'][0]\
                                 ['Ends']['EndEntry'][1]['Value'], 3.8)
            self.assertEqual(procedure['MaccorTestProcedure']['ProcSteps']['TestStep'][0]\
                                 ['Ends']['EndEntry'][2]['Value'], 2.0)
            procedure.to_file(os.path.join(scratch_dir, procedure_to_convert))

            sdu_test_input = os.path.join(SCHEDULE_TEMPLATE_DIR, '20170630-3_6C_9per_5C.sdu')
            converted_sdu_name = 'schedule_test_20200724.sdu'
            proc_dict = procedure.from_file(os.path.join(scratch_dir, procedure_to_convert))

            sdu_test_output = os.path.join(TEST_FILE_DIR, converted_sdu_name)
            test_step_dict = proc_dict['MaccorTestProcedure']['ProcSteps']['TestStep']

            converter = ProcedureToSchedule(test_step_dict)
            global_min_cur = -2 * 1.5 * protocol_params['capacity_nominal']
            global_max_cur = 2 * 1.5 * protocol_params['capacity_nominal']
            converter.create_sdu(sdu_test_input, sdu_test_output, current_range='Range2',
                                 global_v_range=[2.0, 3.8], global_temp_range=[0, 60],
                                 global_current_range=[global_min_cur, global_max_cur])
            parsed = open(sdu_test_output, encoding='latin-1').readlines()
            self.assertEqual(parsed[328], '[Schedule_Step3_Limit0]\n')
            self.assertEqual(parsed[6557], '[Schedule_UserDefineSafety15]\n')
            schedule = Schedule.from_file(os.path.join(sdu_test_output))
            self.assertEqual(schedule['Schedule']['Step15']['m_uLimitNum'], '2')
            self.assertEqual(schedule['Schedule']['Step14']['m_uLimitNum'], '6')
            self.assertEqual(schedule['Schedule']['m_uStepNum'], '96')
            self.assertEqual(schedule['Schedule']['Step86']['m_szCtrlValue'], '15')
            self.assertEqual(schedule['Schedule']['Step86']['m_szExtCtrlValue1'], '1')
            self.assertEqual(schedule['Schedule']['Step86']['m_szExtCtrlValue2'], '0')
            #
            # shutil.copyfile(os.path.join(TEST_FILE_DIR, converted_sdu_name),
            #                 os.path.join(scratch_dir, converted_sdu_name))

    def test_from_csv(self):
        csv_file_list = os.path.join(TEST_FILE_DIR,
                                     "data-share",
                                     "raw",
                                     "parameters",
                                     "Iris_parameters - GP.csv")

        with ScratchDir('.') as scratch_dir:
            makedirs_p(os.path.join(scratch_dir, "procedures"))
            makedirs_p(os.path.join(scratch_dir, "schedules"))
            makedirs_p(os.path.join(scratch_dir, "names"))
            generate_protocol_files_from_csv(csv_file_list, output_directory=scratch_dir)
            if os.path.isfile(os.path.join(scratch_dir, "procedures", ".DS_Store")):
                os.remove(os.path.join(scratch_dir, "procedures", ".DS_Store"))

            procedure_list = os.listdir(os.path.join(scratch_dir, "procedures"))
            for procedure_name in procedure_list:
                sdu_test_input = os.path.join(SCHEDULE_TEMPLATE_DIR, "20170630-3_6C_9per_5C.sdu")
                converted_sdu_name = procedure_name.split(".")[0] + ".sdu"
                proc_dict = Procedure.from_file(os.path.join(scratch_dir, "procedures", procedure_name))
                proc_dict.set_skip_to_end_diagnostic("4.20", "2.9", step_key="070")

                sdu_test_output = os.path.join(scratch_dir, "schedules", converted_sdu_name)
                test_step_dict = proc_dict["MaccorTestProcedure"]["ProcSteps"]["TestStep"]

                converter = ProcedureToSchedule(test_step_dict)
                global_min_cur = -2 * 1.3 * 20
                global_max_cur = 2 * 1.3 * 20
                converter.create_sdu(sdu_test_input, sdu_test_output, current_range="Parallel-High",
                                     global_v_range=[2.85, 4.25], global_temp_range=[0, 55],
                                     global_current_range=[global_min_cur, global_max_cur])

                # # Saving locally
                # shutil.copyfile(os.path.join(scratch_dir, sdu_test_output),
                #                 os.path.join(TEST_FILE_DIR, converted_sdu_name))

            self.assertEqual(len(os.listdir(os.path.join(scratch_dir, "procedures"))), 5)
            self.assertEqual(len(os.listdir(os.path.join(scratch_dir, "schedules"))), 5)

    def test_from_csv_4(self):
        csv_file_list = os.path.join(TEST_FILE_DIR,
                                     "data-share",
                                     "raw",
                                     "parameters",
                                     "Talos_parameters - GP.csv")

        with ScratchDir('.') as scratch_dir:
            makedirs_p(os.path.join(scratch_dir, "procedures"))
            makedirs_p(os.path.join(scratch_dir, "schedules"))
            makedirs_p(os.path.join(scratch_dir, "names"))
            generate_protocol_files_from_csv(csv_file_list, output_directory=scratch_dir)
            if os.path.isfile(os.path.join(scratch_dir, "procedures", ".DS_Store")):
                os.remove(os.path.join(scratch_dir, "procedures", ".DS_Store"))

            procedure_list = os.listdir(os.path.join(scratch_dir, "procedures"))
            for procedure_name in procedure_list:
                sdu_test_input = os.path.join(SCHEDULE_TEMPLATE_DIR, "20170630-3_6C_9per_5C.sdu")
                converted_sdu_name = procedure_name.split(".")[0] + ".sdu"
                proc_dict = Procedure.from_file(os.path.join(scratch_dir, "procedures", procedure_name))
                proc_dict.set_skip_to_end_diagnostic("4.4", "2.1", step_key="070")

                sdu_test_output = os.path.join(scratch_dir, "schedules", converted_sdu_name)
                test_step_dict = proc_dict["MaccorTestProcedure"]["ProcSteps"]["TestStep"]

                converter = ProcedureToSchedule(test_step_dict)
                global_min_cur = -2 * 1.5 * 4
                global_max_cur = 2 * 1.5 * 4
                converter.create_sdu(sdu_test_input, sdu_test_output, current_range="Range1",
                                     global_v_range=[2.0, 4.5], global_temp_range=[0, 60],
                                     global_current_range=[global_min_cur, global_max_cur])
            self.assertEqual(len(os.listdir(os.path.join(scratch_dir, "procedures"))), 35)
            self.assertEqual(len(os.listdir(os.path.join(scratch_dir, "schedules"))), 35)


class ArbinScheduleTest(unittest.TestCase):
    def setUp(self):
        pass

    def test_dict_to_file(self):
        filename = "20170630-3_6C_9per_5C.sdu"
        schedule = Schedule.from_file(os.path.join(SCHEDULE_TEMPLATE_DIR, filename))
        testname = "test1.sdu"
        with ScratchDir("."):
            dumpfn(schedule, "schedule_test.json")
            schedule.to_file(testname)
            hash1 = hash_file(os.path.join(SCHEDULE_TEMPLATE_DIR, filename))
            hash2 = hash_file(testname)
            if hash1 != hash2:
                original = open(
                    os.path.join(SCHEDULE_TEMPLATE_DIR, filename), encoding="latin-1"
                ).read()
                parsed = open(testname, encoding="latin-1").read()
                self.assertFalse(list(difflib.unified_diff(original, parsed)))
                for line in difflib.unified_diff(original, parsed):
                    print(line)

    def test_fastcharge(self):
        filename = "20170630-3_6C_9per_5C.sdu"
        test_file = "test.sdu"
        sdu = Schedule.from_fast_charge(
            1.1 * 3.6, 0.086, 1.1 * 5, os.path.join(SCHEDULE_TEMPLATE_DIR, filename)
        )
        with ScratchDir("."):
            sdu.to_file(test_file)
            hash1 = hash_file(os.path.join(SCHEDULE_TEMPLATE_DIR, filename))
            hash2 = hash_file(test_file)
            if hash1 != hash2:
                original = open(
                    os.path.join(SCHEDULE_TEMPLATE_DIR, filename), encoding="latin-1"
                ).readlines()
                parsed = open(test_file, encoding="latin-1").readlines()
                udiff = list(difflib.unified_diff(original, parsed))
                for line in udiff:
                    print(line)
                self.assertFalse(udiff)


class BiologicSettingsTest(unittest.TestCase):
    def setUp(self):
        pass

    def test_from_file(self):
        filename = "BCS - 171.64.160.115_Ta19_ourprotocol_gdocSEP2019_CC7.mps"
        bcs = Settings.from_file(os.path.join(BIOLOGIC_TEMPLATE_DIR, filename))
        self.assertEqual(len(bcs["Technique"]["1"]["Step"].keys()), 55)
        self.assertEqual(bcs.get_path("Technique.1.Step.5.Ns"), "4")
        self.assertEqual(bcs["Technique"]["1"]["Type"], "Modulo Bat")
        self.assertEqual(bcs["Metadata"]["BT-LAB SETTING FILE"], None)
        self.assertEqual(bcs["Metadata"]["line3"], "blank")
        self.assertEqual(bcs["Metadata"]["Device"], "BCS-805")

    def test_to_file(self):
        filename = "BCS - 171.64.160.115_Ta19_ourprotocol_gdocSEP2019_CC7.mps"
        bcs = Settings.from_file(os.path.join(BIOLOGIC_TEMPLATE_DIR, filename))
        test_name = "test.mps"
        with ScratchDir("."):
            bcs.to_file(test_name)
            original = open(
                os.path.join(BIOLOGIC_TEMPLATE_DIR, filename), encoding="ISO-8859-1"
            ).readlines()
            parsed = open(test_name, encoding="ISO-8859-1").readlines()
            udiff = list(difflib.unified_diff(original, parsed))
            for line in udiff:
                print(line)
            self.assertFalse(udiff)

    def test_insertion(self):
        filename = "BCS - 171.64.160.115_Ta19_ourprotocol_gdocSEP2019_CC7.mps"
        bcs = Settings.from_file(os.path.join(BIOLOGIC_TEMPLATE_DIR, filename))
        value = "{:.3f}".format(5)
        bcs.set("Technique.1.Step.3.ctrl1_val", value)
        self.assertEqual(bcs.get_path("Technique.1.Step.3.ctrl1_val"), "5.000")
        test_name = "test.mps"
        with ScratchDir("."):
            bcs.to_file(test_name)
            original = open(
                os.path.join(BIOLOGIC_TEMPLATE_DIR, filename), encoding="ISO-8859-1"
            ).readlines()
            parsed = open(test_name, encoding="ISO-8859-1").readlines()
            udiff = list(difflib.unified_diff(original, parsed))
            for line in udiff:
                print(line)
            self.assertTrue(udiff)  # Assert that file is not the same as the template

    def test_parameterization(self):
        filename = "formationV1.mps"
        bcs = Settings.from_file(os.path.join(BIOLOGIC_TEMPLATE_DIR, filename))
        protocol_params_df = pd.read_csv(os.path.join(TEST_FILE_DIR,
                                                      "data-share",
                                                      "raw",
                                                      "parameters",
                                                      "Form_parameters - GP.csv"))
        test_name = "test.mps"
        with ScratchDir(".") as scratch_dir:
            makedirs_p(os.path.join(scratch_dir, "settings"))
            for index, protocol_params in protocol_params_df.iterrows():
                template = protocol_params["template"]
                filename_prefix = "_".join(
                    [
                        protocol_params["project_name"],
                        "{:06d}".format(protocol_params["seq_num"]),
                    ]
                )
                if template == "formationV1.mps":
                    bcs = Settings.from_file(os.path.join(BIOLOGIC_TEMPLATE_DIR, filename))

                    self.assertEqual(bcs.get_path("Metadata.Cycle Definition"), "Charge/Discharge alternance")
                    bcs = bcs.formation_protocol_bcs(protocol_params)
                    self.assertEqual(bcs.get_path("Technique.1.Step.2.ctrl1_val"), float(round(0.2 * 0.1, 3)))
                    self.assertEqual(bcs.get_path("Technique.1.Step.3.lim1_value"), float(round(60, 3)))
                    self.assertEqual(bcs.get_path("Technique.1.Step.4.lim1_value"), float(round(30, 3)))
                    self.assertEqual(bcs.get_path("Technique.1.Step.5.ctrl1_val"), float(round(0.2 * 0.2, 3)))
                    self.assertEqual(bcs.get_path("Technique.1.Step.6.lim1_value"), float(round(30, 3)))
                    self.assertEqual(bcs.get_path("Technique.1.Step.7.lim1_value"), float(round(30, 3)))
                    self.assertEqual(bcs.get_path("Technique.1.Step.8.ctrl1_val"), float(round(0.2 * 0.2, 3)))
                    self.assertEqual(bcs.get_path("Technique.1.Step.8.lim1_value"), float(round(3.0, 3)))
                    self.assertEqual(bcs.get_path("Technique.1.Step.9.ctrl1_val"), float(round(0.2 * 0.5, 3)))
                    self.assertEqual(bcs.get_path("Technique.1.Step.10.lim1_value"), float(round(3.9, 3)))
                    self.assertEqual(bcs.get_path("Technique.1.Step.11.ctrl_repeat"), int(1))
                    self.assertEqual(bcs.get_path("Technique.1.Step.12.ctrl1_val"), float(round(0.2 * 1, 3)))
                    self.assertEqual(bcs.get_path("Technique.1.Step.13.lim1_value"), float(round(3.5, 3)))
                    self.assertEqual(bcs.get_path("Technique.1.Step.14.ctrl1_val"), float(round(3.5, 3)))

                test_name = "{}.mps".format(filename_prefix)
                test_name = os.path.join(scratch_dir, "settings", test_name)
                bcs.to_file(test_name)
            self.assertEqual(len(os.listdir(os.path.join(scratch_dir, "settings"))), 16)
            original = open(
                os.path.join(BIOLOGIC_TEMPLATE_DIR, filename), encoding="ISO-8859-1"
            ).readlines()
            parsed = open(test_name, encoding="ISO-8859-1").readlines()
            udiff = list(difflib.unified_diff(original, parsed))
            for line in udiff:
                print(line)
            self.assertTrue(udiff)  # Assert that last file is not the same as the template



"""Classes and functions for handling Indigo battery cycler data.

"""
from datetime import datetime

import pytz
import pandas as pd
import h5py
import numpy as np

from beep.structure.base import BEEPDatapath
from beep.conversion_schemas import INDIGO_CONFIG


class IndigoDatapath(BEEPDatapath):
    """Datapath for ingesting and structuring Indigo battery cycler data.

    """

    @classmethod
    def from_file(cls, path):
        """Creates a BEEPDatapath from an raw Indigo data file.

        Args:
            path (str, Pathlike): file path to data file

        Returns:
            (IndigoDatapath)
        """
        #        if using python 3.7 or 3.8
        #        data = pd.read_hdf(path, "time_series_data")
        #        if using python 3.7, 3.8 or 3.9
        hdf = h5py.File(path, "r")
        d0 = hdf.get('time_series_data')
        d1 = np.array(d0.get('axis0'))
        d2 = np.array(d0.get('axis1'))
        # d3 = np.array(d0.get('block0_items'))
        d4 = np.array(d0.get('block0_values'))
        # d5 = np.array(d0.get('block1_items'))
        d6 = np.array(d0.get('block1_values'))
        # d7 = np.array(d0.get('block2_items'))
        d8 = np.array(d0.get('block2_values'))
        d9 = np.concatenate((d4, d6, d8), axis=1)
        d10 = pd.DataFrame(d9)
        d11 = d10.set_axis(d2, axis="index")
        d1_string = []
        for i in d1:
            j = i.decode("utf-8")
            d1_string.append(j)
        data = d11.set_axis(d1_string, axis="columns")
        metadata = dict()

        if len(list(data["cell_id"].unique())) > 1:
            raise ValueError("More than 1 cell_id exists in {}".format(path))

        metadata["indigo_cell_id"] = int(data["cell_id"].iloc[0])
        metadata["filename"] = path
        metadata["_today_datetime"] = datetime.now().strftime("%Y-%m-%d %H:%M:%S")

        # transformations
        data = (
            data.reset_index().reset_index()
        )  # twice in case old index is stored in file
        data = data.drop(columns=["index"])
        data.rename(columns={"level_0": "data_point"}, inplace=True)
        data.loc[data.half_cycle_count % 2 == 1, "charge_capacity"] = (
                abs(data.cell_coulomb_count_c) / 3600
        )
        data.loc[data.half_cycle_count % 2 == 0, "charge_capacity"] = 0
        data.loc[data.half_cycle_count % 2 == 0, "discharge_capacity"] = (
                abs(data.cell_coulomb_count_c) / 3600
        )
        data.loc[data.half_cycle_count % 2 == 1, "discharge_capacity"] = 0
        data.loc[data.half_cycle_count % 2 == 1, "charge_energy"] = abs(
            data.cell_energy_j
        )
        data.loc[data.half_cycle_count % 2 == 0, "charge_energy"] = 0
        data.loc[data.half_cycle_count % 2 == 0, "discharge_energy"] = abs(
            data.cell_energy_j
        )
        data.loc[data.half_cycle_count % 2 == 1, "discharge_energy"] = 0
        data["internal_resistance"] = data.cell_voltage_v / data.cell_current_a
        data["date_time_iso"] = data["system_time_us"].apply(
            lambda x: datetime.utcfromtimestamp(x / 1000000)
            .replace(tzinfo=pytz.UTC)
            .isoformat()
        )

        data.rename(INDIGO_CONFIG["data_columns"], axis="columns", inplace=True)

        metadata["start_datetime"] = data.sort_values(by="system_time_us")[
            "date_time_iso"
        ].iloc[0]

        paths = {
            "raw": path,
            "metadata": path
        }

        return cls(data, metadata, paths)



#  Copyright (c) 2019 Toyota Research Institute
"""
Top-level module for beep.  Put anything that should
be available in the beep namespace here.
"""
import os
import logging
import sys
from functools import partial
import numpy as np
import watchtower
import time
from botocore.exceptions import NoCredentialsError
from tqdm import tqdm as _tqdm

try:
    from io import StringIO
except (ImportError, ModuleNotFoundError):
    from StringIO import StringIO

# Versioning.  The python code version is frequently tagged
# with a commit hash from the repo, which is supplied via
# an environment variable by the integration build procedure
__version__ = "2022.5.2.14"
VERSION_TAG = os.environ.get("BEEP_VERSION_TAG")
if VERSION_TAG is not None:
    __version__ = "-".join([__version__, VERSION_TAG])

# Custom tqdm with optional turnoff from env
tqdm = partial(_tqdm, disable=bool(os.environ.get("TQDM_OFF")))


JSON_LOG_FMT = {"fmt": (
    '{"time": "%(asctime)s", "level": "%(levelname)s", '
    '"process": "%(process)d", '
    '"module": "%(module)s", "func": "%(funcName)s", '
    '"msg": "%(message)s"}'
)
}

HUMAN_LOG_FMT = {
    "fmt": "%(asctime)s %(levelname)-8s %(message)s",
    "datefmt": "%Y-%m-%d %H:%M:%S"
}


# Common locations
MODULE_DIR = os.path.abspath(os.path.dirname(__file__))
CONVERSION_SCHEMA_DIR = os.path.join(MODULE_DIR, "conversion_schemas")
FEATURES_DIR = os.path.join(MODULE_DIR, "features")
VALIDATION_SCHEMA_DIR = os.path.join(MODULE_DIR, "validation_schemas")
PROTOCOL_PARAMETERS_DIR = os.path.join(MODULE_DIR, "protocol_parameters")
MODEL_DIR = os.path.join(MODULE_DIR, "models")


# All environment variables
BEEP_S3_CACHE_KEY = "BEEP_S3_CACHE"

# Get S3 cache location from env or use default in repo
S3_CACHE = os.environ.get(BEEP_S3_CACHE_KEY, os.path.join(MODULE_DIR, "..", "s3_cache"))

# Logging configuration
LOG_DIR = os.path.join(MODULE_DIR, "logs")
if not os.path.isdir(LOG_DIR):
    os.mkdir(LOG_DIR)
np.set_printoptions(precision=3)
# clear previous loggers/handlers/filters, if exists
logger = logging.getLogger("beep")
logger.handlers = []
logger.filters = []
formatter_stdout = logging.Formatter(**HUMAN_LOG_FMT)
formatter_jsonl = logging.Formatter(**JSON_LOG_FMT)

# Stdout log will always be enabled
hdlr = logging.StreamHandler(sys.stdout)
hdlr.setFormatter(formatter_stdout)
logger.addHandler(hdlr)

logger.setLevel("DEBUG")
logger.propagate = False



# Copyright [2020] [Toyota Research Institute]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Unit tests related to generating waveform files"""

import os
import unittest
import numpy as np
from beep.utils.waveform import convert_velocity_to_power_waveform, RapidChargeWave
import matplotlib.pyplot as plt

from beep.tests.constants import TEST_FILE_DIR


class ChargeWaveformTest(unittest.TestCase):
    def setUp(self):
        pass

    def test_get_input_current_soc_as_x(self):
        charging_c_rates = [0.5, 2.5, 2.0, 0.2]
        soc_points = [0.05, 0.25, 0.65, 0.8]
        final_c_rate = charging_c_rates[-1]
        soc_initial = soc_points[0]
        soc_final = soc_points[-1]
        max_c_rate = 3.0
        min_c_rate = 0.2

        charging = RapidChargeWave(final_c_rate, soc_initial, soc_final, max_c_rate, min_c_rate)

        current_multistep_soc_as_x, soc_vector = charging.get_input_current_multistep_soc_as_x(charging_c_rates,
                                                                                               soc_points)
        self.assertEqual(np.round(np.mean(current_multistep_soc_as_x), 6), np.round(1.8648, 6))
        self.assertEqual(np.round(np.median(current_multistep_soc_as_x), 6), np.round(2.5, 6))
        self.assertEqual(np.round(np.max(current_multistep_soc_as_x), 6), np.round(2.5, 6))
        self.assertEqual(np.round(np.min(current_multistep_soc_as_x), 6), np.round(0.25, 6))

        current_smooth_soc_as_x, soc_vector = charging.get_input_current_smooth_soc_as_x(charging_c_rates, soc_points)
        self.assertEqual(np.round(np.mean(current_smooth_soc_as_x), 6), np.round(1.631819, 6))
        self.assertEqual(np.round(np.median(current_smooth_soc_as_x), 6), np.round(1.991661, 6))
        self.assertEqual(np.round(np.max(current_smooth_soc_as_x), 6), np.round(2.5, 6))
        self.assertEqual(np.round(np.min(current_smooth_soc_as_x), 6), np.round(0.2, 6))

        plt.figure()
        plt.plot(soc_vector, current_smooth_soc_as_x)
        plt.plot(soc_vector, current_multistep_soc_as_x, linestyle='--')
        plt.xlim(0, 1.05)
        plt.ylim(0, 3)
        plt.xlabel('SOC')
        plt.ylabel('C rate [h$^{-1}$]')
        plt.legend(['Smooth', 'Multistep CC', 'CC'])
        plt.savefig(os.path.join(TEST_FILE_DIR, "rapid_charge_soc.png"))

    def test_get_input_current_time_as_x(self):
        charging_c_rates = [0.5, 2.5, 2.0, 0.2]
        soc_points = [0.05, 0.25, 0.65, 0.8]
        final_c_rate = charging_c_rates[-1]
        soc_initial = soc_points[0]
        soc_final = soc_points[-1]
        max_c_rate = 3.0
        min_c_rate = 0.2

        charging = RapidChargeWave(final_c_rate, soc_initial, soc_final, max_c_rate, min_c_rate)

        current_multistep_soc_as_x, soc_vector = charging.get_input_current_multistep_soc_as_x(charging_c_rates,
                                                                                               soc_points)
        time_multistep = charging.get_time_vector_from_c_vs_soc(soc_vector, current_multistep_soc_as_x)
        self.assertEqual(np.round(np.mean(time_multistep), 6), np.round(1552.953655, 6))
        self.assertEqual(np.round(np.median(time_multistep), 6), np.round(1701.081081, 6))
        self.assertEqual(np.round(np.max(time_multistep), 3), 2296.358)

        current_smooth_soc_as_x, soc_vector = charging.get_input_current_smooth_soc_as_x(charging_c_rates, soc_points)
        time_smooth = charging.get_time_vector_from_c_vs_soc(soc_vector, current_smooth_soc_as_x)
        self.assertEqual(np.round(np.mean(time_smooth), 6), np.round(1503.457129, 6))
        self.assertEqual(np.round(np.median(time_smooth), 6), np.round(1644.678275, 6))
        self.assertEqual(np.round(np.max(time_smooth), 3), 2539.486)

        plt.figure()
        plt.plot(time_smooth, current_smooth_soc_as_x)
        plt.plot(time_multistep, current_multistep_soc_as_x, linestyle='--')
        # plt.xlim(0, 1.05)
        plt.ylim(0, 3)
        plt.xlabel('Time [sec]')
        plt.ylabel('C rate [h$^{-1}$]')
        plt.legend(['Smooth', 'Multistep CC'])
        plt.savefig(os.path.join(TEST_FILE_DIR, "rapid_charge_time.png"))

    def test_get_input_current_matching_time(self):
        charging_c_rates = [0.5, 2.5, 2.0, 0.2]
        soc_points = [0.05, 0.25, 0.65, 0.8]
        final_c_rate = charging_c_rates[-1]
        soc_initial = soc_points[0]
        soc_final = soc_points[-1]
        max_c_rate = 3.0
        min_c_rate = 0.2

        charging = RapidChargeWave(final_c_rate, soc_initial, soc_final, max_c_rate, min_c_rate)

        current_smooth, time_smooth, current_multistep, time_multistep = \
            charging.get_input_currents_both_to_final_soc(charging_c_rates, soc_points)

        plt.figure()
        plt.plot(time_smooth, current_smooth)
        plt.plot(time_multistep, current_multistep, linestyle='--')

        plt.ylim(0, 3.5)
        plt.xlabel('Time [sec]')
        plt.ylabel('C rate [h$^{-1}$]')
        plt.legend(['Smooth', 'Multistep CC'])
        plt.savefig(os.path.join(TEST_FILE_DIR, "rapid_charge_matching.png"))

        self.assertEqual(np.round(np.max(time_smooth), 3), np.round(np.max(time_multistep), 3))
        self.assertLessEqual(np.max(current_smooth), max_c_rate)

    def test_dropping_current_matching_time(self):
        charging_c_rates = [2.5, 0.5, 2.0, 0.2]
        soc_points = [0.05, 0.25, 0.65, 0.8]
        final_c_rate = charging_c_rates[-1]
        soc_initial = soc_points[0]
        soc_final = soc_points[-1]
        max_c_rate = 3.0
        min_c_rate = 0.2

        charging = RapidChargeWave(final_c_rate, soc_initial, soc_final, max_c_rate, min_c_rate)

        current_smooth, time_smooth, current_multistep, time_multistep = \
            charging.get_input_currents_both_to_final_soc(charging_c_rates, soc_points)

        plt.figure()
        plt.plot(time_smooth, current_smooth)
        plt.plot(time_multistep, current_multistep, linestyle='--')

        plt.ylim(0, 4)
        plt.xlabel('Time [sec]')
        plt.ylabel('C rate [h$^{-1}$]')
        plt.legend(['Smooth', 'Multistep CC'])
        plt.savefig(os.path.join(TEST_FILE_DIR, "rapid_charge_matching_dropping.png"))

        self.assertLessEqual(np.max(current_smooth), 3)
        self.assertEqual(np.round(np.max(time_smooth), 3), np.round(np.max(time_multistep), 3))

    def test_get_input_current_uniform_time(self):
        charging_c_rates = [0.5, 2.5, 2.0, 0.2]
        soc_points = [0.05, 0.25, 0.65, 0.8]
        final_c_rate = charging_c_rates[-1]
        soc_initial = soc_points[0]
        soc_final = soc_points[-1]
        max_c_rate = 3.0
        min_c_rate = 0.2

        charging = RapidChargeWave(final_c_rate, soc_initial, soc_final, max_c_rate, min_c_rate)
        current_smooth, current_multi, time_uniform = charging.get_currents_with_uniform_time_basis(charging_c_rates,
                                                                                                    soc_points)
        soc_smooth = np.sum(current_smooth) / 3600
        soc_multistep = np.sum(current_multi) / 3600

        plt.figure()
        plt.plot(time_uniform, current_smooth)
        plt.plot(time_uniform, current_multi, linestyle='--')

        plt.ylim(0, 4)
        plt.xlabel('Time [sec]')
        plt.ylabel('C rate [h$^{-1}$]')
        plt.legend(['Smooth', 'Multistep CC'])
        plt.savefig(os.path.join(TEST_FILE_DIR, "rapid_charge_uniform.png"))

        self.assertEqual(np.round(soc_smooth + soc_initial, 2), soc_final)
        self.assertEqual(np.round(soc_multistep + soc_initial, 2), soc_final)
        self.assertTrue(np.all(np.diff(time_uniform) == 1))



"""
Module for S3 operations
"""
import boto3


def download_s3_object(bucket, key, destination_path):
    """
    Download an object from a bucket/key location on S3 to local disk.

    Args:
        bucket (str):                   S3 bucket
        key (str):                      S3 key
        destination_path (str):         local path at which file object is to be stored
    """
    s3_client = boto3.client('s3')
    s3_client.download_file(bucket, key, destination_path)
    return


def list_s3_objects(bucket):
    """
    List all s3 objects available to user on S3 in a specific bucket.

    Args:
        bucket (str): S3 bucket

    Returns:
        ([boto3.ObjectSummary]): List of s3 objects.
    """
    s3 = boto3.resource('s3')
    bucket = s3.Bucket(bucket)
    return list(bucket.objects.all())



"""Classes and functions for handling Neware battery cycler data.

"""

import numpy as np
import pandas as pd
from monty.tempfile import ScratchDir

from beep.structure.base import BEEPDatapath
from beep.structure.maccor import MaccorDatapath
from beep.conversion_schemas import NEWARE_CONFIG


class NewareDatapath(BEEPDatapath):
    """A BEEPDatapath for ingesting and structuring Neware data files.
    """

    @classmethod
    def from_file(cls, filename):
        """Create a NewareDatapath from a raw Neware cycler file.

        Args:
            filename (str, Pathlike): file path for neware file.

        Returns:
            (NewareDatapath)
        """
        ir_column_name = '"DCIR(O)"'
        with open(filename, encoding="ISO-8859-1") as input:
            with ScratchDir("."):
                cycle_header = input.readline().replace("\t", "")
                cycle_file = open("cycle_file.csv", "a", encoding="ISO-8859-1")
                encoded_string = cycle_header.encode("ascii", "ignore")
                cycle_header = encoded_string.decode()
                cycle_file.write(cycle_header)

                step_header = input.readline().replace("\t", "")
                ir_index = step_header.split(",").index(ir_column_name)
                step_file = open("step_file.csv", "a", encoding="ISO-8859-1")
                encoded_string = step_header.encode("ascii", "ignore")
                step_header = encoded_string.decode()
                step_file.write(step_header)

                record_header = input.readline().replace("\t", "")
                record_header = record_header.split(",")
                record_header[0] = cycle_header.split(",")[0]
                record_header[1] = step_header.split(",")[1]
                record_header[22] = ir_column_name
                record_header = ",".join(record_header)
                record_file = open("record_file.csv", "a", encoding="ISO-8859-1")
                encoded_string = record_header.encode("ascii", "ignore")
                record_header = encoded_string.decode()
                record_file.write(record_header)

                # Read file line by line and write to the appropriate file
                cycle_number = 0
                step_number = 0
                for row, line in enumerate(input):
                    if line[:2] == r',"':
                        step_file.write(line)
                        step_number = line.split(",")[1]
                        ir_value = line.split(",")[ir_index]
                    elif line[:2] == r",,":
                        line_list = line.split(",")
                        line_list[0] = cycle_number
                        line_list[1] = step_number
                        line_list[22] = ir_value
                        line = ",".join(line_list)
                        record_file.write(line)
                    else:
                        cycle_file.write(line)
                        cycle_number = line.split(",")[0]
                record_file.close()
                step_file.close()
                cycle_file.close()

                # Read in the data and convert the column values to MKS units
                data = pd.read_csv(
                    "record_file.csv", sep=",", skiprows=0, encoding="ISO-8859-1"
                )
                data = data.loc[:, ~data.columns.str.contains("Unnamed")]
                data["Time(h:min:s.ms)"] = data["Time(h:min:s.ms)"].apply(
                    cls.step_time
                )
                data["Current(mA)"] = data["Current(mA)"] / 1000
                data["Capacitance_Chg(mAh)"] = data["Capacitance_Chg(mAh)"] / 1000
                data["Capacitance_DChg(mAh)"] = data["Capacitance_DChg(mAh)"] / 1000
                data["Engy_Chg(mWh)"] = data["Engy_Chg(mWh)"] / 1000
                data["Engy_DChg(mWh)"] = data["Engy_DChg(mWh)"] / 1000

                # Deal with missing data in the internal resistance
                data["DCIR(O)"] = data["DCIR(O)"].apply(
                    lambda x: np.nan if x == "\t-" else x
                )
                data["DCIR(O)"] = data["DCIR(O)"].fillna(method="ffill")
                data["DCIR(O)"] = data["DCIR(O)"].fillna(method="bfill")

        data["test_time"] = (
            data["Time(h:min:s.ms)"]
            .diff()
            .fillna(0)
            .apply(lambda x: 0 if x < 0 else x)
            .cumsum()
        )
        # print(data.columns)
        # print(NEWARE_CONFIG["data_types"])
        data = data.astype(NEWARE_CONFIG["data_types"])

        data.rename(NEWARE_CONFIG["data_columns"], axis="columns", inplace=True)
        data["date_time"] = data["date_time"].apply(lambda x: x.replace("\t", ""))
        data["date_time_iso"] = data["date_time"].apply(MaccorDatapath.correct_timestamp)

        metadata = dict()
        path = filename

        paths = {
            "raw": path,
            "metadata": path
        }

        return cls(data, metadata, paths)

    @staticmethod
    def step_time(x):
        """Helper function to convert the step time format from Neware h:min:s.ms into
        decimal seconds

        Args:
            x (str): The datetime string for neware in format 'h:min:s.ms'

        Returns:
            float: The time in seconds
        """
        time_list = x.split(":")
        time = (
            3600 * float(time_list[-3]) + 60 * float(time_list[-2]) + float(time_list[-1])
        )
        return time



from beep.structure.base import BEEPDatapath
from beep.structure.arbin import ArbinDatapath
from beep.structure.indigo import IndigoDatapath
from beep.structure.maccor import MaccorDatapath
from beep.structure.neware import NewareDatapath
from beep.structure.biologic import BiologicDatapath
from beep.structure.battery_archive import BatteryArchiveDatapath



# Copyright [2020] [Toyota Research Institute]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Unit tests related to cycler run data structures"""

import os
import unittest
import numpy as np
import pandas as pd

from monty.serialization import loadfn, dumpfn
from monty.tempfile import ScratchDir

from beep.structure.base import BEEPDatapath, step_is_waveform_dchg, step_is_waveform_chg
from beep.structure.arbin import ArbinDatapath
from beep.structure.maccor import MaccorDatapath
from beep.structure.neware import NewareDatapath
from beep.structure.indigo import IndigoDatapath
from beep.structure.biologic import BiologicDatapath, get_cycle_index
from beep.structure.battery_archive import BatteryArchiveDatapath
from beep.tests.constants import TEST_FILE_DIR


class TestArbinDatapath(unittest.TestCase):
    """
    Tests specific to Arbin cyclers.
    """
    def setUp(self) -> None:
        self.bad_file = os.path.join(
            TEST_FILE_DIR, "2017-05-09_test-TC-contact_CH33.csv"
        )

        self.good_file = os.path.join(
            TEST_FILE_DIR, "2017-12-04_4_65C-69per_6C_CH29.csv"
        )

        self.broken_file = os.path.join(
            TEST_FILE_DIR, "Talos_001385_NCR18650618003_CH33_truncated.csv"

        )

    # from RCRT.test_serialization
    def test_serialization(self):
        smaller_run = ArbinDatapath.from_file(self.bad_file)
        with ScratchDir("."):
            dumpfn(smaller_run, "smaller_cycler_run.json")
            resurrected = loadfn("smaller_cycler_run.json")
            self.assertIsInstance(resurrected, BEEPDatapath)
            self.assertIsInstance(resurrected.raw_data, pd.DataFrame)
            self.assertEqual(
                smaller_run.raw_data.voltage.to_list(), resurrected.raw_data.voltage.to_list()
            )
            self.assertEqual(
                smaller_run.raw_data.current.to_list(), resurrected.raw_data.current.to_list()
            )

    def test_from_file(self):
        ad = ArbinDatapath.from_file(self.good_file)
        self.assertEqual(ad.paths.get("raw"), self.good_file)
        self.assertEqual(ad.paths.get("metadata"), self.good_file.replace(".csv", "_Metadata.csv"))
        self.assertTupleEqual(ad.raw_data.shape, (251263, 15))

    # based on PCRT.test_from_arbin_insufficient_interpolation_length
    def test_from_arbin_insufficient_interpolation_length(self):
        rcycler_run = ArbinDatapath.from_file(self.broken_file)
        vrange, num_points, nominal_capacity, fast_charge, diag = rcycler_run.determine_structuring_parameters()
        # print(diag['parameter_set'])
        self.assertEqual(diag['parameter_set'], 'NCR18650-618')
        diag_interp = rcycler_run.interpolate_diagnostic_cycles(diag, resolution=1000, v_resolution=0.0005)
        self.assertAlmostEqual(diag_interp[(diag_interp.cycle_index == 1) &
                                           (diag_interp.step_index == 5)].charge_capacity.max(),
                               3.39608899, 3)
        self.assertAlmostEqual(diag_interp[(diag_interp.cycle_type == "hppc")].charge_capacity.max(),
                               3.4919972, 3)


class TestMaccorDatapath(unittest.TestCase):
    """
    Tests specific to Maccor cyclers.
    """

    def setUp(self) -> None:
        self.good_file = os.path.join(TEST_FILE_DIR, "xTESLADIAG_000019_CH70.070")
        self.diagnostics_file = os.path.join(TEST_FILE_DIR, "xTESLADIAG_000020_CH71.071")
        self.waveform_file = os.path.join(TEST_FILE_DIR, "test_drive_071620.095")
        self.broken_file = os.path.join(TEST_FILE_DIR, "PreDiag_000229_000229_truncated.034")
        self.timezone_file = os.path.join(TEST_FILE_DIR, "PredictionDiagnostics_000109_tztest.010")
        self.timestamp_file = os.path.join(TEST_FILE_DIR, "PredictionDiagnostics_000151_test.052")

        # Note: this is a legacy file
        self.diagnostic_interpolation_file = os.path.join(TEST_FILE_DIR,
                                                          "PredictionDiagnostics_000132_00004C_structure.json")

    # based on RCRT.test_ingestion_maccor
    # based on RCRT.test_timezone_maccor
    # based on RCRT.test_timestamp_maccor
    def test_from_file(self):

        for file in (self.good_file, self.timestamp_file, self.timezone_file):
            md = MaccorDatapath.from_file(file)
            self.assertEqual(md.paths.get("raw"), file)
            self.assertEqual(md.paths.get("metadata"), file)

            if file == self.good_file:
                self.assertTupleEqual(md.raw_data.shape, (11669, 40))
                self.assertEqual(70, md.metadata.channel_id)
            elif file == self.timestamp_file:
                self.assertTupleEqual(md.raw_data.shape, (333, 44))
                self.assertEqual(52, md.metadata.channel_id)
            else:
                self.assertTupleEqual(md.raw_data.shape, (2165, 44))
                self.assertEqual(10, md.metadata.channel_id)

            self.assertEqual(
                set(md.metadata.raw.keys()),
                {
                    "barcode",
                    "_today_datetime",
                    "start_datetime",
                    "filename",
                    "protocol",
                    "channel_id",
                },
            )

            # Quick test to see whether columns get recasted
            self.assertTrue(
                {
                    "data_point",
                    "cycle_index",
                    "step_index",
                    "voltage",
                    "current",
                    "charge_capacity",
                    "discharge_capacity",
                }
                < set(md.raw_data.columns)
            )

    # based on RCRT.test_quantity_sum_maccor
    def test_get_quantity_sum(self):
        md = MaccorDatapath.from_file(self.diagnostics_file)

        cycle_sign = np.sign(np.diff(md.raw_data["cycle_index"]))
        capacity_sign = np.sign(np.diff(md.raw_data["charge_capacity"]))
        self.assertTrue(
            np.all(capacity_sign >= -cycle_sign)
        )  # Capacity increases throughout cycle
        capacity_sign = np.sign(np.diff(md.raw_data["discharge_capacity"]))
        self.assertTrue(
            np.all(capacity_sign >= -cycle_sign)
        )  # Capacity increases throughout cycle

    # based on RCRT.test_whether_step_is_waveform
    def test_step_is_waveform(self):
        md = MaccorDatapath.from_file(self.waveform_file)
        df = md.raw_data
        self.assertTrue(df.loc[df.cycle_index == 6].
                        groupby("step_index").apply(step_is_waveform_dchg).any())
        self.assertFalse(df.loc[df.cycle_index == 6].
                        groupby("step_index").apply(step_is_waveform_chg).any())
        self.assertFalse(df.loc[df.cycle_index == 3].
                        groupby("step_index").apply(step_is_waveform_dchg).any())

    # based on RCRT.test_get_interpolated_waveform_discharge_cycles
    def test_interpolate_waveform_discharge_cycles(self):
        md = MaccorDatapath.from_file(self.waveform_file)
        all_interpolated = md.interpolate_cycles()
        all_interpolated = all_interpolated[(all_interpolated.step_type == "discharge")]
        self.assertTrue(all_interpolated.columns[0] == 'test_time')
        subset_interpolated = all_interpolated[all_interpolated.cycle_index==6]

        df = md.raw_data
        self.assertEqual(subset_interpolated.test_time.min(),
                         df.loc[(df.cycle_index == 6) &
                                (df.step_index == 33)].test_time.min())
        self.assertEqual(subset_interpolated[subset_interpolated.cycle_index == 6].shape[0], 1000)

    # based on RCRT.test_waveform_charge_discharge_capacity
    def test_waveform_charge_discharge_capacity(self):
        md = MaccorDatapath.from_file(self.waveform_file)
        df = md.raw_data
        cycle_sign = np.sign(np.diff(df["cycle_index"]))
        capacity_sign = np.sign(np.diff(df["charge_capacity"]))
        self.assertTrue(
            np.all(capacity_sign >= -cycle_sign)
        )  # Capacity increases throughout cycle
        capacity_sign = np.sign(np.diff(df["discharge_capacity"]))
        self.assertTrue(
            np.all(capacity_sign >= -cycle_sign)
        )

    # based on RCRT.test_get_interpolated_cycles_maccor
    def test_interpolate_cycles(self):
        md = MaccorDatapath.from_file(self.good_file)
        all_interpolated = md.interpolate_cycles(
            v_range=[3.0, 4.2], resolution=10000
        )

        self.assertSetEqual(set(all_interpolated.columns.tolist()),
                            {'voltage',
                             'test_time',
                             'discharge_capacity',
                             'discharge_energy',
                             'current',
                             'charge_capacity',
                             'charge_energy',
                             'internal_resistance',
                             'temperature',
                             'cycle_index',
                             'step_type'}
                            )
        interp2 = all_interpolated[
            (all_interpolated.cycle_index == 2)
            & (all_interpolated.step_type == "discharge")
            ].sort_values("discharge_capacity")
        interp3 = all_interpolated[
            (all_interpolated.cycle_index == 1)
            & (all_interpolated.step_type == "charge")
            ].sort_values("charge_capacity")

        self.assertTrue(interp3.current.mean() > 0)
        self.assertEqual(len(interp3.voltage), 10000)
        self.assertEqual(interp3.voltage.max(), np.float32(4.100838))
#        self.assertEqual(interp3.voltage.min(), np.float32(3.3334765)) # 3.437705 in python3.9
        np.testing.assert_almost_equal(
            interp3[
                interp3.charge_capacity <= interp3.charge_capacity.median()
                ].current.iloc[0],
            2.423209,
            decimal=6,
        )

        df = md.raw_data
        cycle_2 = df[df["cycle_index"] == 2]
        discharge = cycle_2[cycle_2.step_index == 12]
        discharge = discharge.sort_values("discharge_capacity")

        acceptable_error = 0.01
        acceptable_error_offest = 0.001
        voltages_to_check = [3.3, 3.2, 3.1]
        columns_to_check = [
            "voltage",
            "current",
            "discharge_capacity",
            "charge_capacity",
        ]
        for voltage_check in voltages_to_check:
            closest_interp2_index = interp2.index[
                (interp2["voltage"] - voltage_check).abs().min()
                == (interp2["voltage"] - voltage_check).abs()
                ]
            closest_interp2_match = interp2.loc[closest_interp2_index]
            closest_discharge_index = discharge.index[
                (discharge["voltage"] - voltage_check).abs().min()
                == (discharge["voltage"] - voltage_check).abs()
                ]
            closest_discharge_match = discharge.loc[closest_discharge_index]
            for column_check in columns_to_check:
                off_by = (
                        closest_interp2_match.iloc[0][column_check]
                        - closest_discharge_match.iloc[0][column_check]
                )
                self.assertLessEqual(np.abs(off_by),
                        np.abs(closest_interp2_match.iloc[0][column_check])
                        * acceptable_error
                        + acceptable_error_offest)

    # based on PCRT.test_from_maccor_insufficient_interpolation_length
    def test_from_maccor_insufficient_interpolation_length(self):
        md = MaccorDatapath.from_file(self.broken_file)
        vrange, num_points, nominal_capacity, fast_charge, diag = md.determine_structuring_parameters()
        self.assertEqual(diag['parameter_set'], 'Tesla21700')
        diag_interp = md.interpolate_diagnostic_cycles(diag, resolution=1000, v_resolution=0.0005)
        self.assertEqual(np.around(diag_interp[diag_interp.cycle_index == 1].charge_capacity.median(), 3),
                         np.around(0.6371558214610992, 3))

    # based on EISpectrumTest.test_from_maccor
    # todo: needs testing for the entire maccor object
    def test_eis(self):
        path = os.path.join(TEST_FILE_DIR, "maccor_test_file_4267-66-6519.EDA0001.041")
        d = MaccorDatapath.MaccorEIS.from_file(path)


class TestIndigoDatapath(unittest.TestCase):
    # based on RCRT.test_ingestion_indigo
    def test_from_file(self):
        indigo_file = os.path.join(TEST_FILE_DIR, "indigo_test_sample.h5")
        md = IndigoDatapath.from_file(indigo_file)
        self.assertTrue(
            {
                "data_point",
                "cycle_index",
                "step_index",
                "voltage",
                "temperature",
                "current",
                "charge_capacity",
                "discharge_capacity",
            }
            < set(md.raw_data.columns)
        )

        self.assertEqual(
            set(md.metadata.raw.keys()),
            {"indigo_cell_id", "_today_datetime", "start_datetime", "filename"},
        )


class TestBioLogicDatapath(unittest.TestCase):
    # based on RCRT.test_ingestion_biologic
    def test_from_csv(self):

        biologic_file = os.path.join(
            TEST_FILE_DIR, "raw", "test_loopsnewoutput_MB_CE1_short10k.csv"
        )
        dp = BiologicDatapath.from_file(biologic_file)

        self.assertTrue(
            {
                "cycle_index",
                "step_index",
                "voltage",
                "current",
                "discharge_capacity",
                "charge_capacity",
                "data_point",
                "charge_energy",
                "discharge_energy",
            }
            < set(dp.raw_data.columns),
        )

        self.assertEqual(
            {"_today_datetime", "filename", "barcode", "protocol", "channel_id"},
            set(dp.metadata.raw.keys()),
        )

        dp.structure(v_range=[3.0, 4.4])

        self.assertAlmostEqual(dp.structured_summary["charge_capacity"].tolist()[0], 1.4618750, 6)
        self.assertAlmostEqual(dp.structured_summary["discharge_capacity"].tolist()[0], 2.324598, 6)
        self.assertEqual(dp.structured_summary["date_time_iso"].iloc[0], "2021-05-05T22:36:22.757000+00:00")
        self.assertEqual(dp.structured_summary["date_time_iso"].iloc[1], "2021-05-06T09:44:45.604000+00:00")
        self.assertAlmostEqual(dp.raw_data["test_time"].min(), 0, 3)
        self.assertAlmostEqual(dp.raw_data["test_time"].max(), 102040.77, 3)
        # self.assertAlmostEqual(dp.structured_data["test_time"].min(), 13062.720560, 3)
        self.assertAlmostEqual(dp.structured_data["test_time"].min(), 101.089, 2)
        self.assertAlmostEqual(dp.structured_data["test_time"].max(), 101972.885, 3)

    def test_from_txt(self):
        biologic_file = os.path.join(
            TEST_FILE_DIR, "raw", "test_loopsnewoutput_MB_CE1_short10k.txt"
        )
        dp = BiologicDatapath.from_file(biologic_file)

        self.assertTrue(
            {
                "cycle_index",
                "step_index",
                "voltage",
                "current",
                "discharge_capacity",
                "charge_capacity",
                "data_point",
                "charge_energy",
                "discharge_energy",
            }
            < set(dp.raw_data.columns),
        )

        self.assertEqual(
            {"_today_datetime", "filename", "barcode", "protocol", "channel_id"},
            set(dp.metadata.raw.keys()),
        )
        dp.structure(v_range=[3.0, 4.4])

        self.assertAlmostEqual(dp.structured_summary["charge_capacity"].tolist()[0], 1.4618487, 6)
        self.assertAlmostEqual(dp.structured_summary["discharge_capacity"].tolist()[0], 2.324598, 6)
        self.assertEqual(dp.structured_summary["date_time_iso"].iloc[0], "2021-05-05T22:36:00+00:00")
        self.assertEqual(dp.structured_summary["date_time_iso"].iloc[1], "2021-05-06T09:44:22.848000+00:00")
        self.assertAlmostEqual(dp.raw_data["test_time"].min(), 0, 3)
        self.assertAlmostEqual(dp.raw_data["test_time"].max(), 102240.281, 3)
        #self.assertAlmostEqual(dp.structured_data["test_time"].min(), 13062.997, 3)
        self.assertAlmostEqual(dp.structured_data["test_time"].min(), 101.357, 2)
        self.assertAlmostEqual(dp.structured_data["test_time"].max(), 101972.886, 3)

    def test_from_formation_txt(self):
        biologic_file = os.path.join(
            TEST_FILE_DIR, "raw", "test_FormRegu_000100_CG1_Append_short.txt"
        )
        dp = BiologicDatapath.from_file(biologic_file)

        self.assertTrue(
            {
                "cycle_index",
                "step_index",
                "voltage",
                "current",
                "discharge_capacity",
                "charge_capacity",
                "data_point",
                "charge_energy",
                "discharge_energy",
            }
            < set(dp.raw_data.columns),
        )

        dp.structure(v_range=[1.0, 4.4])

        self.assertAlmostEqual(dp.structured_summary["charge_capacity"].tolist()[0], 0.2673875, 6)
        self.assertAlmostEqual(dp.structured_summary["discharge_capacity"].tolist()[0], 0.2573631, 6)
        self.assertEqual(dp.structured_summary["date_time_iso"].iloc[0], "2022-01-18T22:09:40.640000+00:00")
        self.assertEqual(dp.structured_summary["date_time_iso"].iloc[1], "2022-01-22T09:15:10.020000+00:00")
        self.assertAlmostEqual(dp.raw_data["test_time"].min(), 0, 3)
        self.assertAlmostEqual(dp.raw_data["test_time"].max(), 784864.55, 3)
        self.assertAlmostEqual(dp.structured_data["test_time"].min(), 259220.283, 3)
        self.assertAlmostEqual(dp.structured_data["test_time"].max(), 784853.102, 3)
        self.assertGreater(dp.structured_summary["discharge_capacity"].tolist()[4], 0)
        self.assertGreater(dp.structured_summary["discharge_capacity"].tolist()[20], 0)
        self.assertGreater(dp.structured_summary["discharge_capacity"].tolist()[40], 0)

    def test_add_cycle_index(self):

        biologic_file = os.path.join(
            TEST_FILE_DIR, "raw", "test_loopsnewoutput_MB_CE1_short10k.csv"
        )
        df = pd.read_csv(biologic_file, sep=";")
        ns_list = df["Ns"].tolist()
        loop_list = df["Loop"].tolist()
        biotest_file = os.path.join(TEST_FILE_DIR, "BioTest_000001.000.technique_1_cycle_rules.json")
        cycle_index = get_cycle_index(ns_list, biotest_file, loop_list=loop_list)
        c_i = pd.Series(cycle_index)
        self.assertListEqual([1, 2, 3], c_i.unique().tolist())

    def test_mapping_file(self):
        biologic_file = os.path.join(
            TEST_FILE_DIR, "raw", "test_loopsnewoutput_MB_CE1_short10k.txt"
        )
        biotest_file = os.path.join(TEST_FILE_DIR, "BioTest_000001.000.technique_1_cycle_rules.json")
        dp = BiologicDatapath.from_file(biologic_file, mapping_file=biotest_file)
        self.assertIn("cycle_index", dp.raw_data.columns)
        self.assertListEqual([1, 2, 3], dp.raw_data["cycle_index"].unique().tolist())

class TestNewareDatapath(unittest.TestCase):
    # based on RCRT.test_ingestion_neware
    def test_from_file(self):
        neware_file = os.path.join(TEST_FILE_DIR, "raw", "neware_test.csv")
        md = NewareDatapath.from_file(neware_file)
        self.assertEqual(md.raw_data.columns[22], "internal_resistance")
        self.assertTrue(md.raw_data["test_time"].is_monotonic_increasing)
        summary = md.summarize_cycles(nominal_capacity=4.7, full_fast_charge=0.8)
        self.assertEqual(summary["discharge_capacity"].head(5).round(4).tolist(),
                         [2.4393, 2.4343, 2.4255, 2.4221, 2.4210])
        self.assertEqual(summary[summary["cycle_index"] == 55]["discharge_capacity"].round(4).tolist(),
                         [2.3427])


class TestBatteryArchiveDatapath(unittest.TestCase):
    def test_from_file(self):
        ba_file = os.path.join(TEST_FILE_DIR,
                               "SNL_18650_LFP_15C_0-100_0.5-1C_a_timeseries.csv")
        bd = BatteryArchiveDatapath.from_file(ba_file)

        self.assertEqual(bd.raw_data.columns[-1], "internal_resistance")
        self.assertEqual(bd.raw_data.columns[0], "date_time")
        self.assertTrue(bd.raw_data["test_time"].is_monotonic_increasing)

        summary = bd.summarize_cycles()
        self.assertAlmostEqual(summary["temperature_maximum"].loc[3], 16.832001, places=4)
        self.assertAlmostEqual(summary["charge_duration"].loc[4548], 5773.640137, places=4)


if __name__ == "__main__":
    unittest.main()



import os
import unittest

import pandas as pd
from monty.serialization import loadfn
from numpy.testing import assert_array_equal

from beep.structure.cli import auto_load_processed
from beep.features.base import BEEPFeaturizer, BEEPFeatureMatrix, \
    BEEPFeaturizationError
from beep.tests.constants import TEST_FILE_DIR

THIS_DIR = os.path.dirname(os.path.abspath(__file__))


class TestBEEPFeaturizer(unittest.TestCase):
    class ExampleBEEPFeaturizer(BEEPFeaturizer):

        DEFAULT_HYPERPARAMETERS = {
            "hyperparam_A": 12,
            "hyperparam_B": 13
        }

        def validate(self):
            if self.datapath.diagnostic_summary is not None:
                return True, None
            else:
                return False, "no diagnostic info"

        def create_features(self) -> None:
            self.features = pd.DataFrame(
                {
                    "f1": [1 * self.hyperparameters["hyperparam_A"]],
                    "f2": [2 * self.hyperparameters["hyperparam_B"]],
                    "f3": [3]
                }
            )

    @classmethod
    def setUpClass(cls) -> None:
        cls.dp = auto_load_processed(
            os.path.join(TEST_FILE_DIR,
                         "PredictionDiagnostics_000132_00004C_structure.json")
        )

    def test_core_ops(self):
        # with hyperparameters not specified
        f = self.ExampleBEEPFeaturizer(self.dp, hyperparameters=None)
        val, msg = f.validate()
        self.assertTrue(val)
        self.assertIsNone(msg)

        f.create_features()
        X = f.features
        self.assertIsInstance(X, pd.DataFrame)
        self.assertTupleEqual(X.shape, (1, 3))
        self.assertEqual(X["f1"].iloc[0], 12)
        self.assertEqual(X["f2"].iloc[0], 26)
        self.assertEqual(X["f3"].iloc[0], 3)

        # with hyperparameters specified
        hps = {
            "hyperparam_A": 20,
            "hyperparam_B": 3

        }
        f = self.ExampleBEEPFeaturizer(self.dp, hyperparameters=hps)
        val, msg = f.validate()
        self.assertTrue(val)
        self.assertIsNone(msg)

        f.create_features()
        X = f.features
        self.assertIsInstance(X, pd.DataFrame)
        self.assertTupleEqual(X.shape, (1, 3))
        self.assertEqual(X["f1"].iloc[0], 20)
        self.assertEqual(X["f2"].iloc[0], 6)
        self.assertEqual(X["f3"].iloc[0], 3)

        # with incomplete hyperparams
        hps = {
            "hyperparam_A": 11
        }
        with self.assertRaises(BEEPFeaturizationError):
            self.ExampleBEEPFeaturizer(self.dp, hyperparameters=hps)

    def test_serialization(self):
        f = self.ExampleBEEPFeaturizer(self.dp)
        with self.assertRaises(BEEPFeaturizationError):
            f.as_dict()

        f.create_features()
        d = f.as_dict()
        self.assertEqual(d["features"]["f1"][0], 12)
        self.assertEqual(d["hyperparameters"]["hyperparam_A"], 12)
        self.assertEqual(d["metadata"]["barcode"], "00004C")
        self.assertTrue("paths" in d)

        output_path = os.path.join(THIS_DIR,
                                   "BEEPFeaturizer_test_serialization.json.gz")
        f.to_json_file(output_path)
        f_reloaded = f.from_json_file(output_path)
        self.assertIsInstance(f_reloaded.features, pd.DataFrame)
        self.assertEqual(f_reloaded.hyperparameters["hyperparam_B"], 13)
        self.assertEqual(f_reloaded.metadata["channel_id"], 33)


class TestBEEPFeatureMatrix(unittest.TestCase):
    """Test files are 4 featurizers applied to 2 different files,
    for a total of 4 sets of columns for 2 rows.
    """

    def setUp(self) -> None:
        bfs = []

        bf_dir = os.path.join(TEST_FILE_DIR, "featurizer_serialized_files")

        for fname in os.listdir(bf_dir):
            if fname.endswith("_structure.json"):
                fpath = os.path.join(bf_dir, fname)
                bf = loadfn(fpath)
                bfs.append(bf)

        self.featurizers = bfs

    def test_core_ops(self):
        bfm = BEEPFeatureMatrix(self.featurizers)
        self.assertEqual(len(bfm.featurizers), 8)
        self.assertTupleEqual(bfm.matrix.shape, (2, 179))

        dcc = bfm.matrix[
            "var_discharging_capacity::DiagnosticSummaryStats::" \
            "a2c34891d9e2e10bcf61769d24bad986dba94153df4f23a8b4e5716a9b159053"
        ].to_dict()

        for k, v in dcc.items():
            if "PredictionDiagnostics_000132_00004C_structure.json" in k:
                self.assertAlmostEqual(v, -2.674662, 4)
            elif "PreDiag_000440_0000FB_structure.json" in k:
                self.assertAlmostEqual(v, -3.324690, 4)
            else:
                raise ValueError

    def test_serialization(self):
        bfm = BEEPFeatureMatrix(self.featurizers)

        d = bfm.as_dict()
        self.assertTrue("featurizers" in d)
        self.assertTrue("matrix" in d)

        bfm_path = os.path.join(TEST_FILE_DIR, "bfm.json.gz")
        bfm.to_json_file(bfm_path)

        bfm_reloaded = bfm.from_json_file(bfm_path)
        self.assertIsInstance(bfm_reloaded.matrix, pd.DataFrame)
        assert_array_equal(bfm_reloaded.matrix.values, bfm.matrix.values)
